"""
The remote control app for the server process
"""
import argparse
import logging.handlers
import socket
import sys
from argparse import ArgumentDefaultsHelpFormatter
from struct import pack

from dhcpkit.common.server.logging.config_elements import set_verbosity_logger
from typing import Iterable

logger = logging.getLogger()


class ControlClientError(Exception):
    """
    Base class for DHCPKit Control Client errors
    """


class UnknownCommandError(ControlClientError):
    """
    The server doesn't understand the command we sent
    """


class WrongServerError(ControlClientError):
    """
    The socket we connected to doesn't seem to be a DHCPKit server
    """


class CommunicationError(ControlClientError):
    """
    There was a problem communicating
    """


class DHCPKitControlClient:
    """
    A class for communicating with a DHCPKit DHCPv6 server
    """

    def __init__(self, control_socket: str):
        # Open socket
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVTIMEO, pack('ll', 10, 0))
        self.sock.connect(control_socket)

        # Create a buffer for receiving data into
        self.buffer = b''

        # Make sure we are actually connected
        line = self.receive_line()
        if not line.startswith('DHCPKit '):
            raise WrongServerError("Socket doesn't seem to be for DHCPKit")

    def receive_line(self) -> str:
        """
        Receive one line of output from the server

        :return: The received line
        """
        # Stop if the socket is gone
        if not self.sock:
            raise CommunicationError("Reading from a closed connection")

        while True:
            parts = self.buffer.split(b'\n', maxsplit=1)
            if len(parts) == 2:
                # There is a full line in the buffer, return it
                self.buffer = parts[1]
                return parts[0].decode('utf-8')

            # No full line in the buffer, try to get some more data
            try:
                received = self.sock.recv(1024)
                self.buffer += received
            except OSError:
                raise CommunicationError("No response from server")

            # Nothing received: close connection
            if not received:
                self.sock.close()
                self.sock = None
                return None

    def send_command(self, command: str):
        """
        Send a command to the server

        :param command: The command
        """
        # Stop if the socket is gone
        if not self.sock:
            raise CommunicationError("Writing to a closed connection")

        self.sock.send(command.encode('utf-8') + b"\n")

    def execute_command(self, command: str) -> Iterable[str]:
        """
        Send a command and parse the response

        :param command: The command
        :return: The output
        """
        self.send_command(command)

        while True:
            line = self.receive_line()
            if line is None:
                # No more data, the connection is closed
                return

            if line == 'UNKNOWN':
                raise UnknownCommandError("Server doesn't understand '{}'".format(command))

            elif line.startswith('OK:'):
                # Return the information after the OK: tag
                yield line[3:]
                return

            elif line == 'OK':
                return

            else:
                yield line


def handle_args(args: Iterable[str]):
    """
    Handle the command line arguments.

    :param args: Command line arguments
    :return: The arguments object
    """
    parser = argparse.ArgumentParser(
        description="A remote control utility that allows you to send commands to the DHCPv6 server.",
        formatter_class=ArgumentDefaultsHelpFormatter,
        epilog="Use the command 'help' to see which commands the server supports."
    )

    parser.add_argument("command", action="store",
                        help="The command to send to the server")
    parser.add_argument("-v", "--verbosity", action="count", default=0,
                        help="increase output verbosity")
    parser.add_argument("-c", "--control-socket", action="store", metavar="FILENAME",
                        default="/var/run/ipv6-dhcpd.sock",
                        help="location of domain socket for server control")

    args = parser.parse_args(args)

    return args


def main(args: Iterable[str]) -> int:
    """
    The main program loop

    :param args: Command line arguments
    :return: The program exit code
    """
    # Handle command line arguments
    args = handle_args(args)
    set_verbosity_logger(logger, args.verbosity)

    conn = DHCPKitControlClient(args.control_socket)
    output = conn.execute_command(args.command)
    for line in output:
        print(line)

    output = list(conn.execute_command('quit'))
    if output:
        raise CommunicationError("Unexpected reply from server: {}".format(output[0]))


def run() -> int:
    """
    Run the main program and handle exceptions

    :return: The program exit code
    """
    try:
        # Run the server
        return main(sys.argv[1:])
    except Exception as e:
        logger.critical("Error: {}".format(e))
        return 1


if __name__ == '__main__':
    sys.exit(run())



import logging
import unittest
from ipaddress import IPv6Address
from unittest.mock import call

from dhcpkit.ipv6.duids import LinkLayerTimeDUID
from dhcpkit.ipv6.extensions.prefix_delegation import IAPDOption, STATUS_NOPREFIXAVAIL
from dhcpkit.ipv6.messages import AdvertiseMessage, ReplyMessage, RelayForwardMessage, ConfirmMessage, \
    ClientServerMessage
from dhcpkit.ipv6.options import ClientIdOption, ServerIdOption, IANAOption, StatusCodeOption, STATUS_NOADDRSAVAIL, \
    STATUS_USEMULTICAST, STATUS_NOTONLINK
from dhcpkit.ipv6.server.extension_registry import server_extension_registry
from dhcpkit.ipv6.server.filters.marks.config import MarkedWithFilter
from dhcpkit.ipv6.server.handlers import Handler, UseMulticastError
from dhcpkit.ipv6.server.handlers.ignore import IgnoreRequestHandler
from dhcpkit.ipv6.server.handlers.unicast import ServerUnicastOptionHandler
from dhcpkit.ipv6.server.message_handler import MessageHandler
from dhcpkit.ipv6.server.statistics import StatisticsSet
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle
from tests import DeepCopyMagicMock
from tests.ipv6.messages.test_confirm_message import confirm_message
from tests.ipv6.messages.test_request_message import request_message
from tests.ipv6.messages.test_solicit_message import solicit_message


class DummyMarksHandler(Handler):
    def __init__(self, mark: str):
        self.mark = mark
        super().__init__()

    def pre(self, bundle: TransactionBundle):
        bundle.marks.add('pre-' + self.mark)

    def handle(self, bundle: TransactionBundle):
        bundle.marks.add('handle-' + self.mark)

    def post(self, bundle: TransactionBundle):
        bundle.marks.add('post-' + self.mark)


class BadExceptionHandler(Handler):
    def pre(self, bundle: TransactionBundle):
        if bundle.received_over_multicast:
            raise UseMulticastError("Oops, we shouldn't raise this for multicast requests...")


class DummyExtension:
    @staticmethod
    def create_setup_handlers():
        return [DummyMarksHandler('setup')]

    @staticmethod
    def create_cleanup_handlers():
        return [DummyMarksHandler('cleanup')]


class MessageHandlerTestCase(unittest.TestCase):
    def setUp(self):
        # Add a dummy extensions that modifies the marks
        server_extension_registry['dummy'] = DummyExtension()

        # Some mock objects to use
        self.dummy_handler = DeepCopyMagicMock(spec=Handler)
        unicast_me_filter = MarkedWithFilter(filter_condition='unicast-me',
                                             sub_handlers=[ServerUnicastOptionHandler(
                                                 address=IPv6Address('2001:db8::1')
                                             )])
        ignore_me_filter = MarkedWithFilter(filter_condition='ignore-me', sub_handlers=[IgnoreRequestHandler()])
        reject_me_filter = MarkedWithFilter(filter_condition='reject-me', sub_handlers=[BadExceptionHandler()])

        # Prove to PyCharm that this is really a handler
        self.assertIsInstance(self.dummy_handler, Handler)

        # This is the DUID that is used in the message fixtures
        self.duid = LinkLayerTimeDUID(hardware_type=1, time=488458703, link_layer_address=bytes.fromhex('00137265ca42'))

        # Create some message handlers
        self.message_handler = MessageHandler(server_id=self.duid,
                                              sub_filters=[unicast_me_filter, ignore_me_filter, reject_me_filter],
                                              sub_handlers=[self.dummy_handler],
                                              allow_rapid_commit=False,
                                              rapid_commit_rejections=False)
        self.rapid_message_handler = MessageHandler(server_id=self.duid,
                                                    sub_handlers=[self.dummy_handler],
                                                    allow_rapid_commit=True,
                                                    rapid_commit_rejections=False)
        self.very_rapid_message_handler = MessageHandler(server_id=self.duid,
                                                         sub_handlers=[self.dummy_handler],
                                                         allow_rapid_commit=True,
                                                         rapid_commit_rejections=True)

    def test_worker_init(self):
        self.message_handler.worker_init()
        self.dummy_handler.assert_has_calls([
            call.worker_init()
        ])

    def test_empty_message(self):
        with self.assertLogs(level=logging.WARNING) as cm:
            bundle = TransactionBundle(incoming_message=RelayForwardMessage(),
                                       received_over_multicast=True)
            result = self.message_handler.handle(bundle, StatisticsSet())
            self.assertIsNone(result)

        self.assertEqual(len(cm.output), 1)
        self.assertRegex(cm.output[0], '^WARNING:.*:A server should not receive')

    def test_ignorable_multicast_message(self):
        with self.assertLogs(level=logging.DEBUG) as cm:
            bundle = TransactionBundle(incoming_message=solicit_message,
                                       received_over_multicast=True,
                                       marks=['ignore-me'])
            self.message_handler.handle(bundle, StatisticsSet())
            result = bundle.outgoing_message
            self.assertIsNone(result)

        self.assertEqual(len(cm.output), 3)
        self.assertRegex(cm.output[0], '^DEBUG:.*:Handling SolicitMessage')
        self.assertRegex(cm.output[1], '^INFO:.*:Configured to ignore SolicitMessage')
        self.assertRegex(cm.output[2], '^DEBUG:.*:.*ignoring')

    def test_reject_unicast_message(self):
        with self.assertLogs(level=logging.DEBUG) as cm:
            bundle = TransactionBundle(incoming_message=solicit_message,
                                       received_over_multicast=False)
            self.message_handler.handle(bundle, StatisticsSet())
            result = bundle.outgoing_message
            self.assertIsInstance(result, ReplyMessage)
            self.assertEqual(result.get_option_of_type(StatusCodeOption).status_code, STATUS_USEMULTICAST)

        self.assertEqual(len(cm.output), 3)
        self.assertRegex(cm.output[0], '^DEBUG:.*:Handling SolicitMessage')
        self.assertRegex(cm.output[1], '^INFO:.*:Rejecting unicast SolicitMessage')
        self.assertRegex(cm.output[2], '^DEBUG:.*:.*multicast is required')

    def test_accept_unicast_message(self):
        bundle = TransactionBundle(incoming_message=solicit_message,
                                   received_over_multicast=False,
                                   marks=['unicast-me'])
        self.message_handler.handle(bundle, StatisticsSet())
        result = bundle.outgoing_message
        self.assertIsInstance(result, AdvertiseMessage)
        self.assertIsNone(result.get_option_of_type(StatusCodeOption))

    def test_badly_rejected_multicast_message(self):
        with self.assertLogs(level=logging.DEBUG) as cm:
            bundle = TransactionBundle(incoming_message=solicit_message,
                                       received_over_multicast=True,
                                       marks=['reject-me'])
            self.message_handler.handle(bundle, StatisticsSet())
            result = bundle.outgoing_message
            self.assertIsNone(result)

        self.assertEqual(len(cm.output), 3)
        self.assertRegex(cm.output[0], '^DEBUG:.*:Handling SolicitMessage')
        self.assertRegex(cm.output[1], '^DEBUG:.*:.*multicast is required')
        self.assertRegex(cm.output[2], '^ERROR:.*:Not telling client to use multicast')

    def test_solicit_message(self):
        bundle = TransactionBundle(incoming_message=solicit_message,
                                   received_over_multicast=True,
                                   marks=['one', 'two', 'one'])
        self.message_handler.handle(bundle, StatisticsSet())
        result = bundle.outgoing_message

        self.assertIsInstance(result, AdvertiseMessage)
        self.assertEqual(result.transaction_id, solicit_message.transaction_id)
        self.assertEqual(result.get_option_of_type(ClientIdOption), solicit_message.get_option_of_type(ClientIdOption))
        self.assertEqual(result.get_option_of_type(ServerIdOption).duid, self.duid)
        self.assertEqual(result.get_option_of_type(IANAOption).get_option_of_type(StatusCodeOption).status_code,
                         STATUS_NOADDRSAVAIL)
        self.assertEqual(result.get_option_of_type(IAPDOption).get_option_of_type(StatusCodeOption).status_code,
                         STATUS_NOPREFIXAVAIL)

        # Check if the handlers are called correctly
        for method_name in ['pre', 'handle', 'post']:
            method = getattr(self.dummy_handler, method_name)

            self.assertEqual(method.call_count, 1)
            args, kwargs = method.call_args
            self.assertEqual(len(args), 1)
            self.assertEqual(len(kwargs), 0)
            self.assertIsInstance(args[0], TransactionBundle)

        # Check the types and values at various stages
        # In the pre phase there is no response yet
        bundle = self.dummy_handler.pre.call_args[0][0]
        self.assertEqual(bundle.request, solicit_message)
        self.assertEqual(bundle.incoming_relay_messages, [])
        self.assertEqual(bundle.marks, {'one', 'two',
                                        'pre-setup'})
        self.assertIsNone(bundle.response)
        self.assertIsNone(bundle.outgoing_relay_messages)

        # In the handle phase there is an AdvertiseMessage
        bundle = self.dummy_handler.handle.call_args[0][0]
        self.assertEqual(bundle.request, solicit_message)
        self.assertEqual(bundle.incoming_relay_messages, [])
        self.assertEqual(bundle.marks, {'one', 'two',
                                        'pre-setup', 'pre-cleanup',
                                        'handle-setup'})
        self.assertIsInstance(bundle.response, AdvertiseMessage)
        self.assertEqual(bundle.outgoing_relay_messages, [])

        # In the post phase there is still an AdvertiseMessage (no rapid commit)
        bundle = self.dummy_handler.post.call_args[0][0]
        self.assertEqual(bundle.request, solicit_message)
        self.assertEqual(bundle.incoming_relay_messages, [])
        self.assertEqual(bundle.marks, {'one', 'two',
                                        'pre-setup', 'pre-cleanup',
                                        'handle-setup', 'handle-cleanup',
                                        'post-setup'})
        self.assertIsInstance(bundle.response, AdvertiseMessage)
        self.assertEqual(bundle.outgoing_relay_messages, [])

    def test_rapid_solicit_message(self):
        bundle = TransactionBundle(incoming_message=solicit_message,
                                   received_over_multicast=True,
                                   marks=['one', 'two'])
        self.rapid_message_handler.handle(bundle, StatisticsSet())
        result = bundle.outgoing_message

        self.assertIsInstance(result, AdvertiseMessage)
        self.assertEqual(result.transaction_id, solicit_message.transaction_id)
        self.assertEqual(result.get_option_of_type(ClientIdOption), solicit_message.get_option_of_type(ClientIdOption))
        self.assertEqual(result.get_option_of_type(ServerIdOption).duid, self.duid)
        self.assertEqual(result.get_option_of_type(IANAOption).get_option_of_type(StatusCodeOption).status_code,
                         STATUS_NOADDRSAVAIL)
        self.assertEqual(result.get_option_of_type(IAPDOption).get_option_of_type(StatusCodeOption).status_code,
                         STATUS_NOPREFIXAVAIL)

        # Check if the handlers are called correctly
        for method_name in ['pre', 'handle', 'post']:
            method = getattr(self.dummy_handler, method_name)

            self.assertEqual(method.call_count, 1)
            args, kwargs = method.call_args
            self.assertEqual(len(args), 1)
            self.assertEqual(len(kwargs), 0)
            self.assertIsInstance(args[0], TransactionBundle)

        # Check the types and values at various stages
        # In the pre phase there is no response yet
        bundle = self.dummy_handler.pre.call_args[0][0]
        self.assertEqual(bundle.request, solicit_message)
        self.assertEqual(bundle.incoming_relay_messages, [])
        self.assertEqual(bundle.marks, {'one', 'two',
                                        'pre-setup'})
        self.assertIsNone(bundle.response)
        self.assertIsNone(bundle.outgoing_relay_messages)

        # In the handle phase there is an AdvertiseMessage
        bundle = self.dummy_handler.handle.call_args[0][0]
        self.assertEqual(bundle.request, solicit_message)
        self.assertEqual(bundle.incoming_relay_messages, [])
        self.assertEqual(bundle.marks, {'one', 'two',
                                        'pre-setup', 'pre-cleanup',
                                        'handle-setup'})
        self.assertIsInstance(bundle.response, AdvertiseMessage)
        self.assertEqual(bundle.outgoing_relay_messages, [])

        # In the post phase there is still an AdvertiseMessage (rapid commit, but no rapid commit rejections)
        bundle = self.dummy_handler.post.call_args[0][0]
        self.assertEqual(bundle.request, solicit_message)
        self.assertEqual(bundle.incoming_relay_messages, [])
        self.assertEqual(bundle.marks, {'one', 'two',
                                        'pre-setup', 'pre-cleanup',
                                        'handle-setup', 'handle-cleanup',
                                        'post-setup'})
        self.assertIsInstance(bundle.response, AdvertiseMessage)
        self.assertEqual(bundle.outgoing_relay_messages, [])

    def test_very_rapid_solicit_message(self):
        bundle = TransactionBundle(incoming_message=solicit_message,
                                   received_over_multicast=True,
                                   marks=['one'])
        self.very_rapid_message_handler.handle(bundle, StatisticsSet())
        result = bundle.outgoing_message

        self.assertIsInstance(result, ReplyMessage)
        self.assertEqual(result.transaction_id, solicit_message.transaction_id)
        self.assertEqual(result.get_option_of_type(ClientIdOption), solicit_message.get_option_of_type(ClientIdOption))
        self.assertEqual(result.get_option_of_type(ServerIdOption).duid, self.duid)
        self.assertEqual(result.get_option_of_type(IANAOption).get_option_of_type(StatusCodeOption).status_code,
                         STATUS_NOADDRSAVAIL)
        self.assertEqual(result.get_option_of_type(IAPDOption).get_option_of_type(StatusCodeOption).status_code,
                         STATUS_NOPREFIXAVAIL)

        # Check if the handlers are called correctly
        for method_name in ['pre', 'handle', 'post']:
            method = getattr(self.dummy_handler, method_name)

            self.assertEqual(method.call_count, 1)
            args, kwargs = method.call_args
            self.assertEqual(len(args), 1)
            self.assertEqual(len(kwargs), 0)
            self.assertIsInstance(args[0], TransactionBundle)

        # Check the types and values at various stages
        # In the pre phase there is no response yet
        bundle = self.dummy_handler.pre.call_args[0][0]
        self.assertEqual(bundle.request, solicit_message)
        self.assertEqual(bundle.incoming_relay_messages, [])
        self.assertEqual(bundle.marks, {'one',
                                        'pre-setup'})
        self.assertIsNone(bundle.response)
        self.assertIsNone(bundle.outgoing_relay_messages)

        # In the handle phase there is an AdvertiseMessage
        bundle = self.dummy_handler.handle.call_args[0][0]
        self.assertEqual(bundle.request, solicit_message)
        self.assertEqual(bundle.incoming_relay_messages, [])
        self.assertEqual(bundle.marks, {'one',
                                        'pre-setup', 'pre-cleanup',
                                        'handle-setup'})
        self.assertIsInstance(bundle.response, AdvertiseMessage)
        self.assertEqual(bundle.outgoing_relay_messages, [])

        # In the post phase there is a ReplyMessage(rapid commit rejections)
        bundle = self.dummy_handler.post.call_args[0][0]
        self.assertEqual(bundle.request, solicit_message)
        self.assertEqual(bundle.incoming_relay_messages, [])
        self.assertEqual(bundle.marks, {'one',
                                        'pre-setup', 'pre-cleanup',
                                        'handle-setup', 'handle-cleanup',
                                        'post-setup'})
        self.assertIsInstance(bundle.response, ReplyMessage)
        self.assertEqual(bundle.outgoing_relay_messages, [])

    def test_request_message(self):
        bundle = TransactionBundle(incoming_message=request_message,
                                   received_over_multicast=True,
                                   marks=['one'])
        self.message_handler.handle(bundle, StatisticsSet())
        result = bundle.outgoing_message

        self.assertIsInstance(result, ReplyMessage)
        self.assertEqual(result.transaction_id, solicit_message.transaction_id)
        self.assertEqual(result.get_option_of_type(ClientIdOption), solicit_message.get_option_of_type(ClientIdOption))
        self.assertEqual(result.get_option_of_type(ServerIdOption).duid, self.duid)
        self.assertEqual(result.get_option_of_type(IANAOption).get_option_of_type(StatusCodeOption).status_code,
                         STATUS_NOADDRSAVAIL)
        self.assertEqual(result.get_option_of_type(IAPDOption).get_option_of_type(StatusCodeOption).status_code,
                         STATUS_NOPREFIXAVAIL)

    def test_confirm_message(self):
        with self.assertLogs() as cm:
            bundle = TransactionBundle(incoming_message=confirm_message,
                                       received_over_multicast=True,
                                       marks=['one'])
            self.message_handler.handle(bundle, StatisticsSet())
            result = bundle.outgoing_message

        self.assertEqual(len(cm.output), 1)
        self.assertRegex(cm.output[0], '^WARNING:.*:No handler confirmed')

        self.assertIsInstance(result, ReplyMessage)
        self.assertEqual(result.transaction_id, request_message.transaction_id)
        self.assertEqual(result.get_option_of_type(ClientIdOption), solicit_message.get_option_of_type(ClientIdOption))
        self.assertEqual(result.get_option_of_type(ServerIdOption).duid, self.duid)
        self.assertEqual(result.get_option_of_type(StatusCodeOption).status_code, STATUS_NOTONLINK)

    def test_empty_confirm_message(self):
        bundle = TransactionBundle(incoming_message=ConfirmMessage(transaction_id=b'abcd'),
                                   received_over_multicast=True,
                                   marks=['one'])
        self.message_handler.handle(bundle, StatisticsSet())
        result = bundle.outgoing_message

        # ConfirmMessage without IANAOption/IATAOption/IAPDOption must be ignored
        self.assertIsNone(result)

    def test_not_implemented_message(self):
        class NotImplementedMessage(ClientServerMessage):
            message_type = 255
            from_client_to_server = True

        with self.assertLogs() as cm:
            bundle = TransactionBundle(incoming_message=NotImplementedMessage(transaction_id=b'abcd'),
                                       received_over_multicast=True,
                                       marks=['one'])
            self.message_handler.handle(bundle, StatisticsSet())
            result = bundle.outgoing_message

        self.assertEqual(len(cm.output), 1)
        self.assertRegex(cm.output[0], '^WARNING:.*:Do not know how to reply')

        self.assertIsNone(result)


if __name__ == '__main__':
    unittest.main()



"""
Test the InterfaceIdOption implementation
"""
import unittest

from dhcpkit.ipv6.options import InterfaceIdOption
from tests.ipv6.options import test_option


class UnknownOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('0012001030313233343536373839616263646566')
        self.option_object = InterfaceIdOption(b'0123456789abcdef')
        self.parse_option()

    def test_interface_id(self):
        self.option.interface_id = 'BlaBla'
        with self.assertRaisesRegex(ValueError, 'sequence of bytes'):
            self.option.validate()

        self.option.interface_id = 'X' * 2 ** 16
        with self.assertRaisesRegex(ValueError, 'sequence of bytes'):
            self.option.validate()


if __name__ == '__main__':
    unittest.main()



"""
Test the RemoteIdOption implementation
"""
import unittest

from dhcpkit.ipv6.extensions.remote_id import RemoteIdOption
from tests.ipv6.options import test_option


class RemoteIdOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('0025000800009d100123abcd')
        self.option_object = RemoteIdOption(40208, bytes.fromhex('0123abcd'))
        self.parse_option()

    def test_enterprise_number(self):
        self.check_unsigned_integer_property('enterprise_number', size=32)

    def test_remote_id(self):
        self.option.remote_id = 'Not bytes'
        with self.assertRaisesRegex(ValueError, 'sequence of bytes'):
            self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'longer than the available buffer'):
            RemoteIdOption.parse(bytes.fromhex('0025000a00009d10'))


if __name__ == '__main__':
    unittest.main()



"""
Test the DNS options implementations
"""
import unittest
from ipaddress import IPv6Address

from dhcpkit.ipv6.extensions.dns import RecursiveNameServersOption, DomainSearchListOption
from tests.ipv6.options import test_option


class RecursiveNameServersOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('00170020'
                                          '20010db8000000000000000000000001'
                                          '20010db8000000000000000000000002')
        self.option_object = RecursiveNameServersOption(dns_servers=[IPv6Address('2001:db8::1'),
                                                                     IPv6Address('2001:db8::2')])
        self.parse_option()

    def test_validate_addresses(self):
        self.option.dns_servers = IPv6Address('2001:db8::1')
        with self.assertRaisesRegex(ValueError, 'must be a list'):
            self.option.validate()

        self.option.dns_servers = ['2001:db8::1', '2001:db8::2']
        with self.assertRaisesRegex(ValueError, 'IPv6 address'):
            self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'length must be a multiple of 16'):
            RecursiveNameServersOption.parse(bytes.fromhex('0017000f20010db8000000000000000000000001'))

        with self.assertRaisesRegex(ValueError, 'longer than the available buffer'):
            RecursiveNameServersOption.parse(bytes.fromhex('0017001120010db8000000000000000000000001'))

        with self.assertRaisesRegex(ValueError, 'length must be a multiple of 16'):
            RecursiveNameServersOption.parse(bytes.fromhex('0017001120010db800000000000000000000000100'))


class DomainSearchListOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('0018000d') + b'\x08steffann\x02nl\x00'
        self.option_object = DomainSearchListOption(search_list=['steffann.nl'])
        self.parse_option()

    def test_validate_search_list(self):
        self.option.search_list = 'steffann.nl'
        with self.assertRaisesRegex(ValueError, 'must be a list'):
            self.option.validate()

        self.option.search_list = ['steffann.nl', None]
        with self.assertRaisesRegex(ValueError, 'must be a string'):
            self.option.validate()

        self.option.search_list = ['x' + '.x' * 127]
        self.option.validate()

        self.option.search_list = ['xx' + '.x' * 127]
        with self.assertRaisesRegex(ValueError, 'must be 255 characters or less'):
            self.option.validate()

        self.option.search_list = ['www.123456789012345678901234567890123456789012345678901234567890123.nl']
        self.option.validate()

        self.option.search_list = ['www.1234567890123456789012345678901234567890123456789012345678901234.nl']
        with self.assertRaisesRegex(ValueError, 'must be 1 to 63 characters long'):
            self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'must end with a 0-length label'):
            DomainSearchListOption.parse(bytes.fromhex('0018000c') + b'\x08steffann\x02nl\x00')

        with self.assertRaisesRegex(ValueError, 'exceeds available buffer'):
            DomainSearchListOption.parse(bytes.fromhex('0018000e') + b'\x08steffann\x02nl\x00\x01')


if __name__ == '__main__':
    unittest.main()



"""
Base class for pkg_resources based registries
"""
import collections
import logging

import pkg_resources

from dhcpkit.utils import camelcase_to_dash

logger = logging.getLogger(__name__)


class Registry(collections.UserDict):
    """
    Base class for registries
    """

    entry_point = 'dhcpkit.NONE'
    """The name of the entry_point group"""

    def __init__(self):
        """
        A custom dictionary that initialises itself with the entry points from pkg_resources
        """
        super().__init__()

        # A name-based alternative
        self.by_name = {}
        """An alternative name-based mapping"""

        # Try all the entry points
        entry_points = pkg_resources.iter_entry_points(group=self.entry_point)
        for entry_point in entry_points:
            # If the name is a string with an integer then convert it to a real integer
            try:
                name = int(entry_point.name)
            except ValueError:
                name = entry_point.name

            if name in self.data:
                logger.warning("Multiple entry points found for {} {}, using {}".format(
                    self.__class__.__name__, name, self.data[name]))
                continue

            try:
                # Load the entry point and store it
                loaded = entry_point.load()
                self.data[name] = loaded

                # Also store by name
                alternative_name = self.get_name(loaded)
                self.by_name[alternative_name] = loaded
            except pkg_resources.VersionConflict as e:
                # Wrong version, report
                logger.critical("Entry point {} for {} is not compatible: {}".format(
                    entry_point, self.__class__.__name__, e.report()))
                continue
            except ImportError:
                # Ok, this one isn't working, skip it
                logger.error("Entry point {} for {} could not be loaded".format(
                    entry_point, self.__class__.__name__))
                continue

    def get_name(self, item: object) -> str:
        """
        Get the name for the by_name mapping.

        :param item: The item to determine the name of
        :return: The name to use as key in the mapping
        """
        return camelcase_to_dash(item.__name__)



"""
Filters to apply to transaction bundles
"""
import abc
import logging

from cached_property import cached_property
from typing import List, Type, Iterable

from dhcpkit.common.server.config_elements import ConfigElementFactory
from dhcpkit.common.server.logging import DEBUG_HANDLING
from dhcpkit.ipv6.server.handlers import Handler
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle
from dhcpkit.utils import camelcase_to_dash

logger = logging.getLogger(__name__)


class Filter(metaclass=abc.ABCMeta):
    """
    Base class for filters
    """

    def __init__(self, filter_condition: object,
                 sub_filters: Iterable['Filter'] = None, sub_handlers: Iterable[Handler] = None):
        """
        The main initialisation will be done in the master process. After initialisation the master process will create
        worker processes using the multiprocessing module.  Things that can't be pickled and transmitted to the worker
        processes (think database connections etc) have to be initialised separately. Each worker process will call
        worker_init() to do so. Filters that don't need per-worker initialisation can do everything here in __init__().

        :param filter_condition: The condition to filter on
        :param sub_filters: a list of filters configured inside this filter
        :param sub_handlers: a list of handlers configured inside this filter
        """
        self.filter_condition = filter_condition
        self.sub_filters = list(sub_filters or [])
        self.sub_handlers = list(sub_handlers or [])

    def worker_init(self):
        """
        Separate initialisation that will be called in each worker process that is created. Things that can't be forked
        (think database connections etc) have to be initialised here.
        """

        # Cascade to sub-filters and sub-handlers
        for sub_filter in self.sub_filters:
            sub_filter.worker_init()

        for sub_handler in self.sub_handlers:
            sub_handler.worker_init()

    @cached_property
    def filter_description(self) -> str:
        """
        A short description of this filter for log messages.

        :return: The description
        """
        simple_name = camelcase_to_dash(self.__class__.__name__)
        if simple_name.endswith('-filter'):
            simple_name = simple_name[:-7]

        return "{}={}".format(simple_name, self.filter_condition)

    @abc.abstractmethod
    def match(self, bundle: TransactionBundle) -> bool:
        """
        Check whether the given message matches our filter condition.

        :param bundle: The transaction bundle
        :return: Whether our filter condition matches
        """

    def get_handlers(self, bundle: TransactionBundle) -> List[Handler]:
        """
        Get all handlers that are going to be applied to the request in the bundle.

        :param bundle: The transaction bundle
        :return: The list of handlers to apply
        """
        if not self.match(bundle):
            return []

        logger.log(DEBUG_HANDLING, "Filter {} matched".format(self.filter_description, bundle))

        # Collect handlers
        handlers = []

        # Apply all sub-filters and collect their handlers. The idea behind this is that handlers on more-specific
        # filters take precedence over handlers on the outer filters.
        for sub_filter in self.sub_filters:
            handlers += sub_filter.get_handlers(bundle)

        # Now add our own handlers
        handlers += self.sub_handlers

        return handlers


class FilterFactory(ConfigElementFactory, metaclass=abc.ABCMeta):
    """
    Base class for filter factories
    """

    @property
    @abc.abstractmethod
    def filter_class(self) -> Type[Filter]:
        """
        Get the class of filter to create

        :return: The class of filter
        """

    def create(self):
        """
        Create the filter and feed it with the sub-filters and sub-handlers.

        :return: The filter
        """
        sub_filters = []
        for filter_factory in self.filter_factories:
            sub_filters.append(filter_factory())

        sub_handlers = []
        for handler_factory in self.handler_factories:
            sub_handlers.append(handler_factory())

        return self.filter_class(self.filter_condition, sub_filters, sub_handlers)

    @property
    def filter_condition(self):
        """
        Return the filter condition, the name of the section by default
        :return: The filter condition
        """
        return self.name



"""
Extra datatypes for the server configuration
"""

# noinspection PyUnresolvedReferences
import datetime
from ipaddress import IPv6Address

from typing import Type

from dhcpkit.ipv6.messages import Message
from dhcpkit.utils import normalise_hex, camelcase_to_dash

# noinspection PyUnresolvedReferences
from dhcpkit.common.server.config_datatypes import *


def number_of_workers(value: str) -> int:
    """
    The number of worker processes, must be 1 or more
    :param value: The number of processes
    :return: The validated number of processes
    """
    value = int(value)
    if value < 1:
        raise ValueError("Number of workers must be at least 1")
    return value


def unicast_address(value: str) -> IPv6Address:
    """
    Parse an IPv6 address and make sure it is a unicast address

    :param value: The address as string
    :return: The parsed IPv6 address
    """
    address = IPv6Address(value)
    if address.is_link_local or address.is_loopback or address.is_multicast or address.is_unspecified:
        raise ValueError("Address must be a routable IPv6 address")
    return address


def hex_bytes(value: str) -> bytes:
    """
    A sequence of bytes provided as a hexadecimal string.

    :param value: The hexadecimal string
    :return: The corresponding bytes
    """
    value = normalise_hex(value)
    return bytes.fromhex(value)


def iso8601_timestamp(value: str) -> int:
    """
    Convert an ISO8601 timestamp into a unix timestamp

    :param value: ISO8601 formatted timestamp
    :return: Unix timestamp
    """
    dt = datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M:%SZ')
    dt = dt.replace(tzinfo=datetime.timezone.utc)
    return int(dt.timestamp())


def unsigned_int_8(value: str) -> int:
    """
    Parse value as an integer and verify that it is an unsigned 8-bit value.

    :param value: The number as a string
    :return: The corresponding integer
    """
    value = int(value)
    if not (0 <= value < 2 ** 8):
        raise ValueError("The specified value must be an unsigned 8-bit integer")
    return value


def unsigned_int_16(value: str) -> int:
    """
    Parse value as an integer and verify that it is an unsigned 16-bit value.

    :param value: The number as a string
    :return: The corresponding integer
    """
    value = int(value)
    if not (0 <= value < 2 ** 16):
        raise ValueError("The specified value must be an unsigned 16-bit integer")
    return value


def unsigned_int_32(value: str) -> int:
    """
    Parse value as an integer and verify that it is an unsigned 32-bit value.

    :param value: The number as a string
    :return: The corresponding integer
    """
    value = int(value)
    if not (0 <= value < 2 ** 32):
        raise ValueError("The specified value must be an unsigned 32-bit integer")
    return value


def message_type(value: str) -> Type[Message]:
    """
    Parse the value as the name of a DHCPv6 message type

    :param value: The name of the message type
    :return: The message class
    """
    from dhcpkit.ipv6.message_registry import message_registry

    # Prepare the value
    search_value = camelcase_to_dash(value)

    try:
        return message_registry.by_name[search_value]
    except KeyError:
        raise ValueError("{} is not a valid message type".format(value))



"""
Basic handlers for relay options
"""
import logging

from dhcpkit.ipv6.messages import RelayForwardMessage, RelayReplyMessage
from dhcpkit.ipv6.server.handlers import Handler, RelayHandler
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle

logger = logging.getLogger(__name__)


class CopyRelayOptionHandler(RelayHandler):
    """
    This handler just copies a type of option from the incoming relay messages to the outgoing relay messages

    :param option_class: The option class to copy
    """

    def __init__(self, option_class: type(Handler)):
        super().__init__()
        self.option_class = option_class
        """The class of the option from the :class:`.RelayForwardMessage` to the :class:`.RelayReplyMessage`"""

    def handle_relay(self, bundle: TransactionBundle,
                     relay_message_in: RelayForwardMessage, relay_message_out: RelayReplyMessage):
        """
        Copy the options for each relay message pair.

        :param bundle: The transaction bundle
        :param relay_message_in: The incoming relay message
        :param relay_message_out: Thr outgoing relay message
        """
        # Make sure this option can go into this type of response
        if not relay_message_out.may_contain(self.option_class):
            return

        # Make sure this option isn't present and then copy those from the request
        relay_message_out.options = [existing_option for existing_option in relay_message_out.options
                                     if not isinstance(existing_option, self.option_class)]
        relay_message_out.options[:0] = relay_message_in.get_options_of_type(self.option_class)



"""
Handlers for the options defined in dhcpkit.ipv6.extensions.dns
"""

from ipaddress import IPv6Address

from typing import Iterable

from dhcpkit.ipv6.extensions.dns import RecursiveNameServersOption, DomainSearchListOption
from dhcpkit.ipv6.server.handlers.basic import SimpleOptionHandler


class RecursiveNameServersOptionHandler(SimpleOptionHandler):
    """
    Handler for putting RecursiveNameServersOption in responses
    """

    def __init__(self, dns_servers: Iterable[IPv6Address], always_send: bool = False):
        option = RecursiveNameServersOption(dns_servers=dns_servers)
        option.validate()

        super().__init__(option, always_send=always_send)

    def __str__(self):
        return "{} with {}".format(self.__class__.__name__, ', '.join(map(str, self.option.dns_servers)))

    def combine(self, existing_options: Iterable[RecursiveNameServersOption]) -> RecursiveNameServersOption:
        """
        Combine multiple options into one.

        :param existing_options: The existing options to include name servers from
        :return: The combined option
        """
        addresses = []

        # Add from existing options first
        for option in existing_options:
            for address in option.dns_servers:
                if address not in addresses:
                    addresses.append(address)

        # Then add our own
        for address in self.option.dns_servers:
            if address not in addresses:
                addresses.append(address)

        # And return a new option with the combined addresses
        return RecursiveNameServersOption(dns_servers=addresses)


class DomainSearchListOptionHandler(SimpleOptionHandler):
    """
    Handler for putting RecursiveNameServersOption in responses
    """

    def __init__(self, search_list: Iterable[str], always_send: bool = False):
        option = DomainSearchListOption(search_list=search_list)
        option.validate()

        super().__init__(option, always_send=always_send)

    def __str__(self):
        return "{} with {}".format(self.__class__.__name__, ', '.join(self.option.search_list))

    def combine(self, existing_options: Iterable[DomainSearchListOption]) -> DomainSearchListOption:
        """
        Combine multiple options into one.

        :param existing_options: The existing options to include domain names from
        :return: The combined option
        """
        domains = []

        # Add from existing options first
        for option in existing_options:
            for domain in option.search_list:
                if domain not in domains:
                    domains.append(domain)

        # Then add our own
        for domain in self.option.dns_servers:
            if domain not in domains:
                domains.append(domain)

        # And return a new option with the combined addresses
        return DomainSearchListOption(search_list=domains)



"""
Test the Prefix Delegation option implementation
"""
import unittest
from ipaddress import IPv6Network

from dhcpkit.ipv6.extensions.prefix_delegation import IAPrefixOption, IAPDOption
from dhcpkit.ipv6.options import StatusCodeOption, STATUS_NOTONLINK, STATUS_SUCCESS, UnknownOption
from tests.ipv6.options import test_option


class IAPDOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('0019'  # option_type: OPTION_IA_PD
                                          '0045'  # option_length
                                          '41424344'  # iaid: ABCD
                                          '00000029'  # t1: 41
                                          '0000002a'  # t2: 42
                                          '001a'  # option_type: OPTION_IAPREFIX
                                          '0019'  # option_length
                                          '00000000'  # preferred_lifetime
                                          '00000000'  # valid_lifetime
                                          '30'  # prefix_length: 48
                                          '20010db8000000000000000000000000'  # prefix: 2001:db8::
                                          '000d'  # option_type: OPTION_STATUS_CODE
                                          '0018'  # option_length
                                          '0000'  # status_code
                                          '45766572797468696e6720697320617765736f6d6521')  # status_message
        self.option_object = IAPDOption(iaid=b'ABCD', t1=41, t2=42, options=[
            IAPrefixOption(prefix=IPv6Network('2001:db8::/48')),
            StatusCodeOption(status_code=STATUS_SUCCESS, status_message='Everything is awesome!')
        ])
        self.parse_option()

    def test_validate_iaid(self):
        self.option.iaid = b'ABC'
        with self.assertRaisesRegex(ValueError, 'must be four bytes'):
            self.option.validate()

        self.option.iaid = b'ABCDE'
        with self.assertRaisesRegex(ValueError, 'must be four bytes'):
            self.option.validate()

        self.option.iaid = 'ABCD'
        with self.assertRaisesRegex(ValueError, 'must be four bytes'):
            self.option.validate()

    def test_validate_t1(self):
        self.check_unsigned_integer_property('t1', size=32)

    def test_validate_t2(self):
        self.check_unsigned_integer_property('t2', size=32)

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'shorter than the minimum length'):
            IAPDOption.parse(bytes.fromhex('0019000041424344000000290000002a'))

        with self.assertRaisesRegex(ValueError, 'length does not match'):
            IAPDOption.parse(bytes.fromhex('0019000d41424344000000290000002a00140000'))

    def test_sort(self):
        self.assertFalse(self.option > self.option)
        self.assertTrue(self.option <= self.option)

        with self.assertRaises(TypeError):
            self.assertFalse(self.option > 0)

    def test_get_options_of_type(self):
        found_options = self.option.get_options_of_type(StatusCodeOption)
        self.assertEqual(len(found_options), 1)
        self.assertIsInstance(found_options[0], StatusCodeOption)

        # But our test-cases don't have an UnknownOption in them
        found_options = self.option.get_options_of_type(UnknownOption)
        self.assertEqual(len(found_options), 0)

    def test_get_option_of_type(self):
        found_option = self.option.get_option_of_type(StatusCodeOption)
        self.assertIsInstance(found_option, StatusCodeOption)

        # But our test-cases don't have an UnknownOption in them
        found_option = self.option.get_option_of_type(UnknownOption)
        self.assertIsNone(found_option)

    def test_get_prefixes(self):
        prefixes = self.option.get_prefixes()
        self.assertListEqual(prefixes, [IPv6Network('2001:db8::/48')])


class IAPrefixOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('001a0036'
                                          '00015180'
                                          '00093a80'
                                          '30'
                                          '20010db8000100000000000000000000'
                                          '000d0019000457686572652064696420796f752067657420746861743f')
        self.option_object = IAPrefixOption(
            preferred_lifetime=86400,
            valid_lifetime=7 * 86400,
            prefix=IPv6Network('2001:db8:1::/48'),
            options=[
                StatusCodeOption(STATUS_NOTONLINK, 'Where did you get that?')
            ]
        )
        self.parse_option()

    def test_validate_address(self):
        self.option.prefix = '2001:db8::/32'
        with self.assertRaisesRegex(ValueError, 'routable IPv6 prefix'):
            self.option.validate()

        self.option.prefix = IPv6Network('::1/128')
        with self.assertRaisesRegex(ValueError, 'routable IPv6 prefix'):
            self.option.validate()

        self.option.prefix = IPv6Network('fe80::/64')
        with self.assertRaisesRegex(ValueError, 'routable IPv6 prefix'):
            self.option.validate()

        self.option.prefix = IPv6Network('ff02::1/128')
        with self.assertRaisesRegex(ValueError, 'routable IPv6 prefix'):
            self.option.validate()

    def test_validate_preferred_lifetime(self):
        self.check_unsigned_integer_property('preferred_lifetime', size=32)

    def test_validate_valid_lifetime(self):
        self.check_unsigned_integer_property('valid_lifetime', size=32)

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'shorter than the minimum length'):
            IAPrefixOption.parse(bytes.fromhex('001a00090001518000093a8030'))

        with self.assertRaisesRegex(ValueError, 'length does not match'):
            IAPrefixOption.parse(bytes.fromhex('001a0035'
                                               '00015180'
                                               '00093a80'
                                               '30'
                                               '20010db8000100000000000000000000'
                                               '000d0019000457686572652064696420796f752067657420746861743f'))


if __name__ == '__main__':
    unittest.main()



"""
Handlers for the basic :rfc:`3315` options
"""

import logging

from dhcpkit.ipv6.duids import DUID
from dhcpkit.ipv6.options import ServerIdOption
from dhcpkit.ipv6.server.handlers import CannotRespondError
from dhcpkit.ipv6.server.handlers.basic import OverwriteOptionHandler
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle

logger = logging.getLogger(__name__)


class ForOtherServerError(CannotRespondError):
    """
    A specific case of being unable to respond: this message is for another server
    """
    pass


class ServerIdHandler(OverwriteOptionHandler):
    """
    The handler for ServerIdOption. Checks whether any server-id in the request matches our own and puts our server-id
    in the response message to let the client know who is answering.

    :type option: ServerIdOption
    """

    option = None

    def __init__(self, duid: DUID):
        """
        Create a handler function based on the provided DUID

        :param duid: The DUID of this server
        """
        option = ServerIdOption(duid)
        option.validate()

        super().__init__(option, always_send=True)

    def pre(self, bundle: TransactionBundle):
        """
        Check if there is a ServerId in the request

        :param bundle: The transaction bundle
        """
        server_id = bundle.request.get_option_of_type(ServerIdOption)
        if server_id and server_id.duid != self.option.duid:
            # This message is not for this server
            raise ForOtherServerError



"""
Handlers for the options defined in dhcpkit.ipv6.extensions.remote_id
"""
import logging

from dhcpkit.ipv6.extensions.remote_id import RemoteIdOption
from dhcpkit.ipv6.server.handlers.basic_relay import CopyRelayOptionHandler

logger = logging.getLogger(__name__)


class CopyRemoteIdOptionHandler(CopyRelayOptionHandler):
    """
    The handler for RemoteIdOptions in relay messages
    """

    def __init__(self):
        super().__init__(RemoteIdOption)



"""
Handlers for the options defined in dhcpkit.ipv6.extensions.sol_max_rt
"""
from dhcpkit.ipv6.extensions.sol_max_rt import SolMaxRTOption, InfMaxRTOption
from dhcpkit.ipv6.server.handlers.basic import OverwriteOptionHandler


class SolMaxRTOptionHandler(OverwriteOptionHandler):
    """
    Handler for putting SolMaxRTOption in responses
    """

    def __init__(self, sol_max_rt: int, always_send: bool = False):
        option = SolMaxRTOption(sol_max_rt=sol_max_rt)
        option.validate()

        super().__init__(option, always_send=always_send)

    def __str__(self):
        return "{} with {}".format(self.__class__.__name__, self.option.sol_max_rt)


class InfMaxRTOptionHandler(OverwriteOptionHandler):
    """
    Handler for putting InfMaxRTOption in responses
    """

    def __init__(self, inf_max_rt: int, always_send: bool = False):
        option = InfMaxRTOption(inf_max_rt=inf_max_rt)
        option.validate()

        super().__init__(option, always_send=always_send)

    def __str__(self):
        return "{} with {}".format(self.__class__.__name__, self.option.inf_max_rt)



"""
Handlers for the options defined in dhcpkit.ipv6.extensions.sip_serves
"""

from ipaddress import IPv6Address

from typing import Iterable

from dhcpkit.ipv6.extensions.sip_servers import SIPServersDomainNameListOption, SIPServersAddressListOption
from dhcpkit.ipv6.server.handlers.basic import SimpleOptionHandler


class SIPServersDomainNameListOptionHandler(SimpleOptionHandler):
    """
    Handler for putting SIPServersDomainNameListOptions in responses
    """

    def __init__(self, domain_names: Iterable[str], always_send: bool = False):
        option = SIPServersDomainNameListOption(domain_names=domain_names)
        option.validate()

        super().__init__(option, always_send=always_send)

    def __str__(self):
        return "{} with {}".format(self.__class__.__name__, ', '.join(map(str, self.option.domain_names)))

    def combine(self, existing_options: Iterable[SIPServersDomainNameListOption]) -> SIPServersDomainNameListOption:
        """
        Combine multiple options into one.

        :param existing_options: The existing options to include NTP servers from
        :return: The combined option
        """
        domain_names = []

        # Add from existing options first
        for option in existing_options:
            for domain_name in option.domain_names:
                if domain_name not in domain_names:
                    domain_names.append(domain_name)

        # Then add our own
        for domain_name in self.option.domain_names:
            if domain_name not in domain_names:
                domain_names.append(domain_name)

        # And return a new option with the combined addresses
        return SIPServersDomainNameListOption(domain_names=domain_names)


class SIPServersAddressListOptionHandler(SimpleOptionHandler):
    """
    Handler for putting SIPServersAddressListOptions in responses
    """

    def __init__(self, sip_servers: Iterable[IPv6Address], always_send: bool = False):
        option = SIPServersAddressListOption(sip_servers=sip_servers)
        option.validate()

        super().__init__(option, always_send=always_send)

    def __str__(self):
        return "{} with {}".format(self.__class__.__name__, ', '.join(map(str, self.option.sip_servers)))

    def combine(self, existing_options: Iterable[SIPServersAddressListOption]) -> SIPServersAddressListOption:
        """
        Combine multiple options into one.

        :param existing_options: The existing options to include NTP servers from
        :return: The combined option
        """
        sip_servers = []

        # Add from existing options first
        for option in existing_options:
            for sip_server in option.sip_servers:
                if sip_server not in sip_servers:
                    sip_servers.append(sip_server)

        # Then add our own
        for sip_server in self.option.sip_servers:
            if sip_server not in sip_servers:
                sip_servers.append(sip_server)

        # And return a new option with the combined addresses
        return SIPServersAddressListOption(sip_servers=sip_servers)



"""
Test the implementation of option length checking
"""
import unittest

from struct import pack

from dhcpkit.ipv6.option_registry import option_registry
from dhcpkit.ipv6.options import Option
from tests.ipv6.options import test_option


# A dummy option that may not be in a RelayMessageOption
class LengthTestingOption(Option):
    option_type = 65535

    def __init__(self, data: bytes = b''):
        self.data = data

    def load_from(self, buffer: bytes, offset: int = 0, length: int = None) -> int:
        my_offset, option_len = self.parse_option_header(buffer, offset, length, min_length=1, max_length=2)

        self.data = buffer[offset + my_offset:offset + my_offset + option_len]
        my_offset += option_len

        return my_offset

    def save(self) -> bytes:
        return pack('!HH', self.option_type, len(self.data)) + self.data


class RelayMessageOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        option_registry[65535] = LengthTestingOption

        self.option_bytes = bytes.fromhex('ffff000161')
        self.option_object = LengthTestingOption(data=b'a')
        self.parse_option()

    def tearDown(self):
        del option_registry[65535]

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'shorter than the minimum length'):
            LengthTestingOption.parse(bytes.fromhex('ffff0000'))

        with self.assertRaisesRegex(ValueError, 'longer than the maximum length'):
            LengthTestingOption.parse(bytes.fromhex('ffff0003'))


if __name__ == '__main__':
    unittest.main()



"""
Handlers for the basic :rfc:`3315` options
"""

import logging

from dhcpkit.ipv6.options import ClientIdOption
from dhcpkit.ipv6.server.handlers.basic import CopyOptionHandler

logger = logging.getLogger(__name__)


class ClientIdHandler(CopyOptionHandler):
    """
    The handler for ClientIdOptions
    """

    def __init__(self):
        super().__init__(ClientIdOption, always_send=True)



"""
Implementation of a listener on a local multicast network interface
"""
import logging
import netifaces
import socket
from ipaddress import IPv6Address
from struct import pack

from typing import Iterable

from dhcpkit.ipv6 import SERVER_PORT, All_DHCP_Relay_Agents_and_Servers
from dhcpkit.ipv6.server.listeners import Listener, ListenerFactory
from dhcpkit.ipv6.utils import is_global_unicast

logger = logging.getLogger(__name__)


class MulticastInterfaceListenerFactory(ListenerFactory):
    """
    Factory for the implementation of a listener on a local multicast network interface
    """

    name_datatype = staticmethod(str)

    def validate_config_section(self):
        """
        Validate the interface information
        """

        try:
            socket.if_nametoindex(self.name)
        except OSError:
            raise ValueError("Interface {} not found".format(self.name))

        interface_addresses = [IPv6Address(addr_info['addr'].split('%')[0])
                               for addr_info
                               in netifaces.ifaddresses(self.name).get(netifaces.AF_INET6, [])]

        # Pick the first link-local address as reply-from if none is specified in the configuration
        if not self.section.reply_from:
            for address in interface_addresses:
                if address.is_link_local:
                    self.section.reply_from = address
                    break

            if not self.section.reply_from:
                raise ValueError("No link-local address found on interface {}".format(self.name))

        else:
            # Validate what the user supplied
            if not self.section.reply_from.is_link_local:
                raise ValueError("The reply-from address must be a link-local address")

            if self.section.reply_from not in interface_addresses:
                raise ValueError("Cannot find reply-from address {} on interface {}".format(self.section.reply_from,
                                                                                            self.name))

        # Pick the first global unicast address as link-address if none is specified in the configuration
        if not self.section.link_address:
            for address in interface_addresses:
                if is_global_unicast(address):
                    self.section.link_address = address
                    break

            if not self.section.link_address:
                # Use the unspecified address is we couldn't find anything
                self.section.link_address = IPv6Address('::')

        else:
            # Validate what the user supplied (we don't really care if it exists, it's just extra information for the
            # option handlers
            if not is_global_unicast(self.section.link_address):
                raise ValueError("The link-address must be a global unicast address")

    def create(self, old_listeners: Iterable[Listener] = None) -> Listener:
        """
        Create a listener of this class based on the configuration in the config section.

        :param old_listeners: A list of existing listeners in case we can recycle them
        :return: A listener object
        """
        mc_address = IPv6Address(All_DHCP_Relay_Agents_and_Servers)
        interface_index = socket.if_nametoindex(self.name)

        # Try recycling
        old_listeners = list(old_listeners or [])
        for old_listener in old_listeners:
            if self.match_socket(sock=old_listener.listen_socket, address=mc_address, interface=interface_index):
                logger.debug("Recycling existing multicast socket on {}".format(self.name))
                mc_sock = old_listener.listen_socket
                break
        else:
            logger.debug("Listening for multicast requests on {}".format(self.name))
            mc_sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            mc_sock.bind((str(mc_address), SERVER_PORT, 0, interface_index))
            mc_sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_JOIN_GROUP,
                               pack('16sI', mc_address.packed, interface_index))

        # Set the socket options
        mc_sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_LOOP, self.listen_to_self and 1 or 0)

        for old_listener in old_listeners:
            if self.match_socket(sock=old_listener.listen_socket, address=self.reply_from, interface=interface_index):
                logger.debug("  - Recycling existing reply socket for {} on {}".format(self.reply_from, self.name))
                ll_sock = old_listener.listen_socket
                break

            if self.match_socket(sock=old_listener.reply_socket, address=self.reply_from, interface=interface_index):
                logger.debug("  - Recycling existing reply socket for {} on {}".format(self.reply_from, self.name))
                ll_sock = old_listener.reply_socket
                break
        else:
            logger.debug("  - Sending replies from {}".format(self.reply_from))
            ll_sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            ll_sock.bind((str(self.reply_from), SERVER_PORT, 0, interface_index))

        return Listener(interface_name=self.name, listen_socket=mc_sock, reply_socket=ll_sock,
                        global_address=self.link_address, marks=self.marks)



"""
Worker process for handling requests using multiprocessing.
"""
import logging
import logging.handlers
import re
import signal
from multiprocessing import Queue, current_process

from dhcpkit.ipv6 import SERVER_PORT, CLIENT_PORT
from dhcpkit.ipv6.messages import RelayForwardMessage, Message, RelayReplyMessage
from dhcpkit.ipv6.options import InterfaceIdOption, RelayMessageOption
from dhcpkit.ipv6.server.listeners import IncomingPacketBundle, OutgoingPacketBundle
from dhcpkit.ipv6.server.message_handler import MessageHandler
from dhcpkit.ipv6.server.queue_logger import WorkerQueueHandler
from dhcpkit.ipv6.server.statistics import ServerStatistics
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle
from typing import Optional

# These globals will be set by setup_worker()

logger = None
""":type: logging.Logger"""

logging_handler = None
""":type: WorkerQueueHandler"""

current_message_handler = None
""":type: MessageHandler"""

shared_statistics = None
""":type: ServerStatistics"""


def setup_worker(message_handler: MessageHandler, logging_queue: Queue, lowest_log_level: int,
                 statistics: ServerStatistics):
    """
    This function will be called after a new worker process has been created. Its purpose is to set the global
    variables in this specific worker process so that they can be reused across multiple requests. Otherwise we would
    have to pickle them each and every time, and because they are static that would be a waste.

    :param message_handler: The message handler for the incoming requests
    :param logging_queue: The queue where we can deposit log messages so the main process can log them
    :param lowest_log_level: The lowest log level that is going to be handled by the main process
    :param statistics: Container for shared memory with statistics counters
    """
    # Let's shorten the process name a bit by removing everything except the "Worker-x" bit at the end
    this_process = current_process()
    this_process.name = re.sub(r'^.*(Worker-\d+)$', r'\1', this_process.name)

    # Ignore normal signal handling
    signal.signal(signal.SIGINT, lambda signum, frame: None)
    signal.signal(signal.SIGTERM, lambda signum, frame: None)
    signal.signal(signal.SIGHUP, lambda signum, frame: None)

    # Save the logger, don't let it filter, send everything to the queue
    global logger
    logger = logging.getLogger()
    logger.setLevel(logging.NOTSET)

    global logging_handler
    logging_handler = WorkerQueueHandler(logging_queue)
    logging_handler.setLevel(lowest_log_level)
    logger.addHandler(logging_handler)

    # Save the message handler
    global current_message_handler
    current_message_handler = message_handler

    global shared_statistics
    shared_statistics = statistics

    # Run the per-process startup code for the message handler and its children
    message_handler.worker_init()


def parse_incoming_request(incoming_packet: IncomingPacketBundle) -> TransactionBundle:
    """
    Parse the incoming packet and add a RelayServerMessage around it containing the meta-data received from the
    listener.

    :param incoming_packet: The received packet
    :return: The parsed message in a transaction bundle
    """
    length, incoming_message = Message.parse(incoming_packet.data)

    # Determine the next hop count and construct useful log messages
    if isinstance(incoming_message, RelayForwardMessage):
        next_hop_count = incoming_message.hop_count + 1
    else:
        next_hop_count = 0

    # Pretend to be an internal relay and wrap the message like a relay would
    wrapped_message = RelayForwardMessage(hop_count=next_hop_count,
                                          link_address=incoming_packet.link_address,
                                          peer_address=incoming_packet.sender,
                                          options=[
                                              InterfaceIdOption(
                                                  interface_id=incoming_packet.interface_name.encode('utf-8')),
                                              RelayMessageOption(relayed_message=incoming_message)
                                          ])

    # Create the transaction bundle
    return TransactionBundle(incoming_message=wrapped_message,
                             received_over_multicast=incoming_packet.received_over_multicast,
                             marks=incoming_packet.marks)


def generate_outgoing_response(outgoing_message: Message,
                               incoming_packet: IncomingPacketBundle = None) -> OutgoingPacketBundle:
    """
    generate the outgoing packet and check the RelayServerMessage around it.

    :param outgoing_message: The reply message
    :param incoming_packet: The original received packet, only used to sanity-check the reply
    :return: The parsed message
    """
    # Verify that the outer relay message makes sense
    if not isinstance(outgoing_message, RelayReplyMessage):
        raise ValueError("The reply has to be wrapped in a RelayReplyMessage")

    if incoming_packet is not None:
        # Verify the contents of the outgoing message
        if outgoing_message.link_address != incoming_packet.link_address:
            raise ValueError("The relay-reply link-address does not match the relay-forward link-address")

        interface_id_option = outgoing_message.get_option_of_type(InterfaceIdOption)
        if interface_id_option and interface_id_option.interface_id != incoming_packet.interface_name.encode('utf-8'):
            # If there is an interface-id option its contents have to match
            raise ValueError("The interface-id in the reply does not match the interface-id of the request")

    reply = outgoing_message.relayed_message
    if not reply:
        raise ValueError("The RelayReplyMessage does not contain a message")

    # Down to network addresses and bytes
    port = isinstance(reply, RelayReplyMessage) and SERVER_PORT or CLIENT_PORT
    destination = outgoing_message.peer_address
    data = reply.save()

    return OutgoingPacketBundle(message_id=incoming_packet.message_id, data=data, destination=destination, port=port)


def handle_message(incoming_packet: IncomingPacketBundle) -> Optional[OutgoingPacketBundle]:
    """
    Handle a single incoming request. This is supposed to be called in a separate worker thread that has been
    initialised with setup_worker().

    :param incoming_packet: The raw incoming request
    :returns: The packet to reply with and the destination
    """
    # Set the log_id to make it easier to correlate log messages
    logging_handler.log_id = incoming_packet.message_id

    # Until we parsed the packet we can only update global and interface statistics
    statistics = shared_statistics.get_update_set(interface_name=incoming_packet.interface_name)

    try:
        try:
            # Parse the packet
            bundle = parse_incoming_request(incoming_packet)
        except Exception as e:
            logger.error("Error while parsing request: {}".format(e))

            # Count the packet on the statistics counters that we have
            statistics.count_incoming_packet()
            statistics.count_unparsable_packet()
            return

        # Now we know more: update all statistics and count the packet on all
        statistics = shared_statistics.get_update_set(interface_name=incoming_packet.interface_name, bundle=bundle)
        statistics.count_incoming_packet()

        try:
            current_message_handler.handle(bundle, statistics)

            outgoing_message = bundle.outgoing_message
            if outgoing_message:
                out = generate_outgoing_response(outgoing_message, incoming_packet)
                statistics.count_outgoing_packet()
                return out
        except Exception as e:
            logger.error("Error while handling request: {}".format(e))
            statistics.count_handling_error()
            return

    finally:
        # Always reset the log_id when leaving
        logging_handler.log_id = None



"""
Statistics about the server in shared memory
"""
from collections import OrderedDict
from ctypes import c_uint64
from multiprocessing import Value
from multiprocessing.sharedctypes import Synchronized

from dhcpkit.ipv6.message_registry import message_registry
from dhcpkit.ipv6.messages import ClientServerMessage
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle
from dhcpkit.utils import camelcase_to_underscore
from typing import Dict, Hashable, List, Iterable


def create_update_method(counter_name):
    """
    Create a counting method for a simple counter on the Statistics class

    :param counter_name: The name of the counter to update
    :return: The generated method
    """

    def count_method(self):
        """
        Call the counting method on all statistics objects
        """
        counter = getattr(self, counter_name)
        with counter.get_lock():
            counter.value += 1

    return count_method


def create_update_dict_method(counter_name):
    """
    Create a counting method for a counter in a dictionary on the Statistics class

    :param counter_name: The name of the counter to update
    :return: The generated method
    """

    def count_method(self, key):
        """
        Update the counter for the given key
        """
        counter_dict = getattr(self, counter_name)
        if key in counter_dict:
            counter = counter_dict[key]
            with counter.get_lock():
                counter.value += 1

    return count_method


def create_count_method(method_name: str):
    """
    Create a counting method for the StatisticsSet class

    :param method_name: The name of the method to call on Statistics objects
    :return: The generated method
    """

    def count_method(self):
        """
        Call the counting method on all statistics objects
        """
        for stats in self.statistics_set:
            method = getattr(stats, method_name)
            method()

    return count_method


def create_count_dict_method(method_name: str):
    """
    Create a counting method for the StatisticsSet class

    :param method_name: The name of the method to call on Statistics objects
    :return: The generated method
    """

    def count_method(self, key):
        """
        Call the counting method on all statistics objects
        """
        for stats in self.statistics_set:
            method = getattr(stats, method_name)
            method(key)

    return count_method


class Statistics:
    """
    A set of statistics about DHCPv6

    :type incoming_packets: Synchronized
    :type outgoing_packets: Synchronized

    :type unparsable_packets: Synchronized
    :type handling_errors: Synchronized

    :type for_other_server: Synchronized
    :type do_not_respond: Synchronized
    :type use_multicast: Synchronized

    :type messages_in: Dict[int, Synchronized]
    :type messages_out: Dict[int, Synchronized]
    """

    def __init__(self):
        # Packet counts
        self.incoming_packets = Value(c_uint64)
        self.outgoing_packets = Value(c_uint64)

        # Errors
        self.unparsable_packets = Value(c_uint64)
        self.handling_errors = Value(c_uint64)

        # Special replies
        self.for_other_server = Value(c_uint64)
        self.do_not_respond = Value(c_uint64)
        self.use_multicast = Value(c_uint64)

        # Counters per message type
        self.messages_in = OrderedDict()
        self.messages_out = OrderedDict()

        # Which message types do we know about?
        message_registry_keys = list(message_registry.keys())
        message_registry_keys.sort()

        for key in message_registry_keys:
            message_class = message_registry[key]
            if message_class.from_client_to_server and issubclass(message_class, ClientServerMessage):
                self.messages_in[message_class.message_type] = Value(c_uint64)

            if message_class.from_server_to_client and issubclass(message_class, ClientServerMessage):
                self.messages_out[message_class.message_type] = Value(c_uint64)

    def __str__(self):
        lines = [
            "Packets",
            "- Incoming packets: {}".format(self.incoming_packets.value),
            "- Outgoing packets: {}".format(self.outgoing_packets.value),
            "Errors",
            "- Unparsable packets: {}".format(self.unparsable_packets.value),
            "- Handling errors: {}".format(self.handling_errors.value),
            "Special replies",
            "- For other server: {}".format(self.for_other_server.value),
            "- Do not respond: {}".format(self.do_not_respond.value),
            "- Use multicast: {}".format(self.use_multicast.value),
            "Incoming messages",
        ]

        for message_type, counter in self.messages_in.items():
            message_type_name = message_registry[message_type].__name__
            if message_type_name.endswith('Message'):
                message_type_name = message_type_name[:-7]
            lines += ['- {}: {}'.format(message_type_name, counter.value)]

        lines += ['Outgoing messages']

        for message_type, counter in self.messages_out.items():
            message_type_name = message_registry[message_type].__name__
            if message_type_name.endswith('Message'):
                message_type_name = message_type_name[:-7]
            lines += ['- {}: {}'.format(message_type_name, counter.value)]

        return '\n'.join(lines)

    def export(self) -> Dict[str, int]:
        """
        Export the counters

        :return: The counters in a processable format
        """
        out = OrderedDict()
        out['incoming_packets'] = self.incoming_packets.value
        out['outgoing_packets'] = self.outgoing_packets.value
        out['unparsable_packets'] = self.unparsable_packets.value
        out['handling_errors'] = self.handling_errors.value
        out['for_other_server'] = self.for_other_server.value
        out['do_not_respond'] = self.do_not_respond.value
        out['use_multicast'] = self.use_multicast.value

        out['messages_in'] = OrderedDict()
        for message_type, counter in self.messages_in.items():
            message_type_name = message_registry[message_type].__name__
            message_type_name = camelcase_to_underscore(message_type_name)
            if message_type_name.endswith('_message'):
                message_type_name = message_type_name[:-8]
            out['messages_in'][message_type_name] = counter.value

        out['messages_out'] = OrderedDict()
        for message_type, counter in self.messages_out.items():
            message_type_name = message_registry[message_type].__name__
            message_type_name = camelcase_to_underscore(message_type_name)
            if message_type_name.endswith('_message'):
                message_type_name = message_type_name[:-8]
            out['messages_out'][message_type_name] = counter.value

        return out

    count_incoming_packet = create_update_method('incoming_packets')
    count_outgoing_packet = create_update_method('outgoing_packets')
    count_unparsable_packet = create_update_method('unparsable_packets')
    count_handling_error = create_update_method('handling_errors')
    count_for_other_server = create_update_method('for_other_server')
    count_do_not_respond = create_update_method('do_not_respond')
    count_use_multicast = create_update_method('use_multicast')
    count_message_in = create_update_dict_method('messages_in')
    count_message_out = create_update_dict_method('messages_out')


class StatisticsSet:
    """
    A set of statistics objects that are updated together. The metaclass will create all methods for us.
    """

    def __init__(self, statistics_set: Iterable[Statistics]=None):
        self.statistics_set = set(statistics_set or [])

    count_incoming_packet = create_count_method('count_incoming_packet')
    count_outgoing_packet = create_count_method('count_outgoing_packet')
    count_unparsable_packet = create_count_method('count_unparsable_packet')
    count_handling_error = create_count_method('count_handling_error')
    count_for_other_server = create_count_method('count_for_other_server')
    count_do_not_respond = create_count_method('count_do_not_respond')
    count_use_multicast = create_count_method('count_use_multicast')
    count_message_in = create_count_dict_method('count_message_in')
    count_message_out = create_count_dict_method('count_message_out')


class ServerStatistics:
    """
    A set of statistics about the DHCPv6 server

    :type global_stats: Statistics
    :type interface_stats: Dict[str, Statistics]
    :type subnet_stats: Dict[IPv6Network, Statistics]
    :type relay_stats: Dict[IPv6Address, Statistics]
    """

    def __init__(self):
        self.global_stats = Statistics()

        # On-demand categories
        self.interface_stats = {}
        self.subnet_stats = {}
        self.relay_stats = {}

    def set_categories(self, category_settings):
        """
        Create space for the given interfaces

        :param category_settings: Configuration setting for categories
        """
        if not category_settings:
            return

        def update_categories(container: Dict[Hashable, Statistics], categories: Iterable[Hashable]):
            """
            Update a category dictionary based on the provided list of wanted categories

            :param container: The container to update
            :param categories: The list of wanted categories
            """
            # Keep track of existing categories that we don't want anymore
            remaining = set(container.keys())

            # Create categories
            for key in categories:
                if key not in container:
                    container[key] = Statistics()

                    if key in remaining:
                        remaining.remove(key)

            # Remove unwanted categories
            for key in remaining:
                del container[key]

        update_categories(self.interface_stats, category_settings.interfaces)
        update_categories(self.subnet_stats, category_settings.subnets)
        update_categories(self.relay_stats, category_settings.relays)

    def get_update_set(self, interface_name: str = None, bundle: TransactionBundle = None) -> StatisticsSet:
        """
        Return all statistics objects that need to be updated.

        :param interface_name: The name of the interface that we received the packet on
        :param bundle: The transaction bundle to base the selection on
        :return: The set to call count methods on
        """
        stats_set = [self.global_stats]

        if interface_name and interface_name in self.interface_stats:
            stats_set.append(self.interface_stats[interface_name])

        if bundle:
            link_address = bundle.link_address
            for subnet, stats in self.subnet_stats.items():
                if link_address in subnet:
                    stats_set.append(stats)

            relays = bundle.relays
            for address, stats in self.relay_stats.items():
                if address in relays:
                    stats_set.append(stats)

        return StatisticsSet(stats_set)

    def __str__(self):
        lines = ['Global']
        lines += [('- ' if not line.startswith('- ') else '  ') + line
                  for line in str(self.global_stats).split('\n')]

        def get_category_lines(type_name: str, category_data: Dict[Hashable, Statistics]) -> List[str]:
            """
            Get lines for a category

            :param type_name: A descriptive type name for the categories
            :param category_data: A dictionary for categories
            :return: The lines for the provided category
            """
            sub_lines = []

            keys = list(category_data.keys())
            keys.sort()

            for name in keys:
                sub_lines += ['', '{} {}'.format(type_name, name)]
                sub_lines += [('- ' if not line.startswith('- ') else '  ') + line
                              for line in str(category_data[name]).split('\n')]

            return sub_lines

        lines += get_category_lines('Interface', self.interface_stats)
        lines += get_category_lines('Subnet', self.subnet_stats)
        lines += get_category_lines('Relay', self.relay_stats)

        return '\n'.join(lines)

    def export(self) -> Dict[str, int]:
        """
        Export the counters

        :return: The counters in a processable format
        """
        out = OrderedDict()
        out['global'] = self.global_stats.export()

        def get_category_data(category_data: Dict[Hashable, Statistics]) -> Dict[str, dict]:
            """
            Get data for a category

            :param category_data: A dictionary for categories
            :return: The data for the provided categories
            """
            data = OrderedDict()

            keys = list(category_data.keys())
            keys.sort()

            for name in keys:
                data[str(name)] = category_data[name].export()

            return data

        out['interfaces'] = get_category_data(self.interface_stats)
        out['subnets'] = get_category_data(self.subnet_stats)
        out['relays'] = get_category_data(self.relay_stats)

        return out



"""
Test the StatusCodeOption implementation
"""
import unittest

from dhcpkit.ipv6.options import StatusCodeOption, STATUS_NOTONLINK
from tests.ipv6.options import test_option


class StatusCodeOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('000d001d00044fc3b920c3aa7465732d766f7573206d6f6e206d61c3ae7472653f')
        self.option_object = StatusCodeOption(STATUS_NOTONLINK, 'Où êtes-vous mon maître?')
        self.parse_option()

    def test_status_code(self):
        self.check_unsigned_integer_property('status_code', size=16)

    def test_status_message(self):
        self.option.status_message = b'This is not a string'
        with self.assertRaisesRegex(ValueError, 'must be a string'):
            self.option.validate()


if __name__ == '__main__':
    unittest.main()



import unittest

from dhcpkit.ipv6.server.handlers import Handler


class TestHandler(Handler):
    pass


class HandlerTestCase(unittest.TestCase):
    def test_str(self):
        handler = TestHandler()
        self.assertEqual(str(handler), 'TestHandler')


if __name__ == '__main__':
    unittest.main()



"""
Adapt the QueueListener so that it respects the log levels of the handlers. Based on the Python 3.5 implementation.
"""
from logging.handlers import QueueListener, QueueHandler
from multiprocessing.queues import Queue, Full


class QueueLevelListener(QueueListener):
    """
    QueueListener that respects log levels
    """

    def handle(self, record):
        """
        Handle a record.

        This just loops through the handlers offering them the record to handle.
        """
        record = self.prepare(record)
        for handler in self.handlers:
            if record.levelno >= handler.level:
                handler.handle(record)

    # noinspection PyPep8Naming
    def addHandler(self, handler):
        """
        Add the specified handler to this logger.
        """
        if handler not in self.handlers:
            self.handlers.append(handler)

    # noinspection PyPep8Naming
    def removeHandler(self, handler):
        """
        Remove the specified handler from this logger.
        """
        if handler in self.handlers:
            self.handlers.remove(handler)

    def dequeue(self, block):
        """
        Dequeue a record and return it, optionally blocking. Return the sentinel on EOF because otherwise there are
        strange errors after a reload.
        """
        try:
            return self.queue.get(block)
        except EOFError:
            return self._sentinel


class WorkerQueueHandler(QueueHandler):
    """
    A logging handler that queues messages and doesn't cause exceptions when the queue is full.
    """

    def __init__(self, queue: Queue):
        super().__init__(queue)
        self.log_id = None

    def prepare(self, record):
        """
        Prepares a record for queuing. The object returned by this method is
        enqueued. This implementation adds the log_id if it is set.
        """
        record = super().prepare(record)

        # Put in the log_id if it is set
        if self.log_id is not None:
            log_id = str(self.log_id)
            record.message = log_id + ': ' + record.message
            record.msg = record.message

        return record

    def enqueue(self, record):
        """
        Enqueue a record.

        Try three times rapidly, then just drop it.
        """
        for _ in (1, 2, 3):
            try:
                self.queue.put_nowait(record)
                return
            except Full:
                pass



"""
The basic configuration objects
"""

import abc
import logging

from ZConfig.matcher import SectionValue

logger = logging.getLogger(__name__)


class ConfigSection:
    """
    Basic configuration section
    """

    name_datatype = None
    """The datatype of the name of this section. Sections with datatype None cannot have a name"""

    def __init__(self, section: SectionValue):
        self.section = section
        """The `SectionValue` we received as input from the parser"""

        self.name = self.section.getSectionName()
        """The parsed value of the section name"""

        if self.name:
            if not self.name_datatype:
                raise ValueError("{} cannot have a name".format(self.__class__.__name__))

            # Convert and validate
            self.name = self.name_datatype(self.name)

        self.clean_config_section()
        self.validate_config_section()

        if self.name_datatype and not self.name:
            raise ValueError("{} must have {} name".format(self.__class__.__name__, self.name_datatype.__name__))

    # noinspection PyMethodMayBeStatic
    def clean_config_section(self):
        """
        Clean up the config, calculating defaults etc.
        """

    # noinspection PyMethodMayBeStatic
    def validate_config_section(self):
        """
        Validate if the information in the config section is acceptable
        """

    def __getattr__(self, item: str):
        """
        Try to resolve non-existent properties from the raw section data

        :param item: The name of the property
        :return: The value from the section data
        """
        return getattr(self.section, item)

    def __str__(self):
        return self.to_str()

    def to_str(self, indent: int = 0) -> str:
        """
        Return a readable string representation of this element. Because it is not in the real configuration file format
        we don't attempt to make it look like one. We intentionally make it look different.

        :param indent: How much indentation at the start of this element
        :return: The formatted representation
        """
        # Build output in here
        output = ''

        # Section type name, or None for the root
        section_type = self.section.getSectionType()

        # Opening line, unless this is the root
        if section_type:
            name = self.section.getSectionName()
            if name:
                output += '  ' * indent + '{} {}:\n'.format(section_type, name)
            else:
                output += '  ' * indent + '{}:\n'.format(section_type)

            indent += 1

        # Attributes
        attributes = self.section.getSectionAttributes()
        for key in attributes:
            # Get the value, and always make it a list, even if there is only one value, for easier formatting
            value = getattr(self.section, key)
            if isinstance(value, list):
                values = value
            else:
                values = [value]

            for value in values:
                if isinstance(value, ConfigSection):
                    # Config sections can format themselves
                    output += value.to_str(indent)
                else:
                    # For other types we just use their string representation
                    output += '  ' * indent + '{} = {}\n'.format(key, str(value))

        return output


class ConfigElementFactory(ConfigSection, metaclass=abc.ABCMeta):
    """
    Base class for factories to create elements from configuration
    """

    def __init__(self, section: SectionValue):
        self._element = None
        super().__init__(section)

    @abc.abstractmethod
    def create(self, *args, **kwargs) -> object:
        """
        Override this method to create the element.

        :return: The element
        """
        return None

    def __call__(self, *args, **kwargs) -> object:
        """
        Create the handler on demand and return it.

        :return: The option handler
        """
        # Create the handler if we haven't done so yet
        if self._element is None:
            self._element = self.create(*args, **kwargs)

        return self._element



"""
An extension to get static assignments from CSV files, Shelves or an SQLite database
"""
import logging
from abc import ABCMeta, abstractmethod
from collections import namedtuple
from ipaddress import IPv6Network, IPv6Address

from typing import Iterable, Optional

from dhcpkit.common.server.logging import DEBUG_HANDLING
from dhcpkit.ipv6.extensions.prefix_delegation import IAPDOption, IAPrefixOption
from dhcpkit.ipv6.messages import SolicitMessage, RequestMessage, ConfirmMessage, RenewMessage, RebindMessage, \
    ReleaseMessage, DeclineMessage
from dhcpkit.ipv6.options import IANAOption, IAAddressOption
from dhcpkit.ipv6.server.handlers import Handler
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle
from dhcpkit.ipv6.utils import prefix_overlaps_prefixes

logger = logging.getLogger(__name__)

Assignment = namedtuple('Assignment', ['address', 'prefix'])


class StaticAssignmentHandler(Handler, metaclass=ABCMeta):
    """
    An option handler that gives a static address and/or prefix to clients
    """

    def __init__(self,
                 address_preferred_lifetime: int, address_valid_lifetime: int,
                 prefix_preferred_lifetime: int, prefix_valid_lifetime: int):
        """
        Initialise the mapping. This handler will respond to clients on responsible_for_links and assume that all
        addresses in the mapping are appropriate for on those links.

        :param address_preferred_lifetime: The preferred lifetime in seconds for addresses
        :param address_valid_lifetime: The valid lifetime in seconds for addresses
        :param prefix_preferred_lifetime: The preferred lifetime in seconds for prefixes
        :param prefix_valid_lifetime: The valid lifetime in seconds for prefixes
        """
        super().__init__()
        self.address_preferred_lifetime = address_preferred_lifetime
        self.address_valid_lifetime = address_valid_lifetime
        self.prefix_preferred_lifetime = prefix_preferred_lifetime
        self.prefix_valid_lifetime = prefix_valid_lifetime

    @abstractmethod
    def get_assignment(self, bundle: TransactionBundle) -> Assignment:
        """
        Subclasses override this method to determine the assignment for the request in the bundle. This MUST return
        an Assignment object, even if no addresses are provided in it.

        :param bundle: The transaction bundle
        :return: The assignment
        """

    @staticmethod
    def find_iana_option_for_address(options: Iterable[IANAOption], address: IPv6Address) -> Optional[IANAOption]:
        """
        Find an IANAOption that contains the given address

        :param options: The list of options to search
        :param address: The address to look for
        :return: The matching option, if any
        """
        options = list(options)

        for option in options:
            for suboption in option.get_options_of_type(IAAddressOption):
                if suboption.address == address:
                    return option

        # Nothing found: default to first
        return options[0] if options else None

    @staticmethod
    def find_iapd_option_for_prefix(options: Iterable[IAPDOption], prefix: IPv6Network) -> Optional[IAPDOption]:
        """
        Find an IAPDOption that contains the given prefix

        :param options: The list of options to search
        :param prefix: The prefix to look for
        :return: The matching option, if any
        """
        options = list(options)

        for option in options:
            for suboption in option.get_options_of_type(IAPrefixOption):
                if suboption.prefix == prefix:
                    return option

        # Nothing found: default to first
        return options[0] if options else None

    def handle(self, bundle: TransactionBundle):
        """
        The handling is so complex that we just delegate the implementation to separate methods.

        :param bundle: The transaction bundle
        """
        if isinstance(bundle.request, (SolicitMessage, RequestMessage)):
            self.handle_request(bundle)
        elif isinstance(bundle.request, ConfirmMessage):
            self.handle_confirm(bundle)
        elif isinstance(bundle.request, (RenewMessage, RebindMessage)):
            self.handle_renew_rebind(bundle)
        elif isinstance(bundle.request, (ReleaseMessage, DeclineMessage)):
            self.handle_release_decline(bundle)

    def handle_request(self, bundle: TransactionBundle):
        """
        Handle a client requesting addresses (also handles SolicitMessage)

        :param bundle: The request bundle
        """
        # Get the assignment
        assignment = self.get_assignment(bundle)

        # Try to assign the prefix first: it's not dependent on the link
        if assignment.prefix:
            unanswered_iapd_options = bundle.get_unhandled_options(IAPDOption)
            found_option = self.find_iapd_option_for_prefix(unanswered_iapd_options, assignment.prefix)
            if found_option:
                # Answer to this option
                logger.log(DEBUG_HANDLING, "Assigning prefix {}".format(assignment.prefix))
                response_option = IAPDOption(found_option.iaid, options=[
                    IAPrefixOption(prefix=assignment.prefix,
                                   preferred_lifetime=self.prefix_preferred_lifetime,
                                   valid_lifetime=self.prefix_valid_lifetime)
                ])
                bundle.response.options.append(response_option)
                bundle.mark_handled(found_option)
            else:
                logger.log(DEBUG_HANDLING,
                           "Prefix {} reserved, but client did not ask for it".format(assignment.prefix))

        if assignment.address:
            unanswered_iana_options = bundle.get_unhandled_options(IANAOption)
            found_option = self.find_iana_option_for_address(unanswered_iana_options, assignment.address)
            if found_option:
                # Answer to this option
                logger.log(DEBUG_HANDLING, "Assigning address {}".format(assignment.address))
                response_option = IANAOption(found_option.iaid, options=[
                    IAAddressOption(address=assignment.address,
                                    preferred_lifetime=self.address_preferred_lifetime,
                                    valid_lifetime=self.address_valid_lifetime)
                ])
                bundle.response.options.append(response_option)
                bundle.mark_handled(found_option)
            else:
                logger.log(DEBUG_HANDLING,
                           "Address {} reserved, but client did not ask for it".format(assignment.address))

    def handle_confirm(self, bundle: TransactionBundle):
        """
        Handle a client requesting confirmation

        :param bundle: The request bundle
        """
        # Get the assignment
        assignment = self.get_assignment(bundle)

        # Collect unanswered options
        unanswered_iana_options = bundle.get_unhandled_options(IANAOption)

        # See if there are any addresses on a link that I am responsible for
        for option in unanswered_iana_options:
            for suboption in option.get_options_of_type(IAAddressOption):
                if suboption.address == assignment.address:
                    # This is the address from the assignment: it's ok
                    bundle.mark_handled(option)
                    continue

    def handle_renew_rebind(self, bundle: TransactionBundle):
        """
        Handle a client renewing/rebinding addresses

        :param bundle: The request bundle
        """
        # Get the assignment
        assignment = self.get_assignment(bundle)

        # Collect unanswered options
        unanswered_iana_options = bundle.get_unhandled_options(IANAOption)
        unanswered_iapd_options = bundle.get_unhandled_options(IAPDOption)

        for option in unanswered_iapd_options:
            if assignment.prefix and prefix_overlaps_prefixes(assignment.prefix, option.get_prefixes()):
                # Overlap with our assigned prefix: take responsibility
                response_suboptions = []
                for suboption in option.get_options_of_type(IAPrefixOption):
                    if suboption.prefix == assignment.prefix:
                        # This is the correct option, renew it
                        logger.log(DEBUG_HANDLING, "Renewing prefix {}".format(assignment.prefix))
                        response_suboptions.append(IAPrefixOption(prefix=assignment.prefix,
                                                                  preferred_lifetime=self.prefix_preferred_lifetime,
                                                                  valid_lifetime=self.prefix_valid_lifetime))
                    else:
                        # This isn't right
                        logger.log(DEBUG_HANDLING, "Withdrawing prefix {}".format(suboption.prefix))
                        response_suboptions.append(IAPrefixOption(prefix=suboption.prefix,
                                                                  preferred_lifetime=0, valid_lifetime=0))

                response_option = IAPDOption(option.iaid, options=response_suboptions)
                bundle.response.options.append(response_option)
                bundle.mark_handled(option)

        for option in unanswered_iana_options:
            response_suboptions = []
            for suboption in option.get_options_of_type(IAAddressOption):
                if suboption.address == assignment.address:
                    # This is the correct option, renew it
                    logger.log(DEBUG_HANDLING, "Renewing address {}".format(assignment.address))
                    response_suboptions.append(IAAddressOption(address=assignment.address,
                                                               preferred_lifetime=self.address_preferred_lifetime,
                                                               valid_lifetime=self.address_valid_lifetime))
                else:
                    # This isn't right
                    logger.log(DEBUG_HANDLING, "Withdrawing address {}".format(suboption.address))
                    response_suboptions.append(IAAddressOption(address=suboption.address,
                                                               preferred_lifetime=0, valid_lifetime=0))

            response_option = IANAOption(option.iaid, options=response_suboptions)
            bundle.response.options.append(response_option)
            bundle.mark_handled(option)

    def handle_release_decline(self, bundle: TransactionBundle):
        """
        Handle a client releasing or declining resources. Doesn't really need to do anything because assignments are
        static. Just mark the right options as handled.

        :param bundle: The request bundle
        """
        # Get the assignment
        assignment = self.get_assignment(bundle)

        # Collect unanswered options
        unanswered_iana_options = bundle.get_unhandled_options(IANAOption)
        unanswered_iapd_options = bundle.get_unhandled_options(IAPDOption)

        for option in unanswered_iapd_options:
            if assignment.prefix and prefix_overlaps_prefixes(assignment.prefix, option.get_prefixes()):
                # Overlap with our assigned prefix: take responsibility
                bundle.mark_handled(option)

        for option in unanswered_iana_options:
            if assignment.address in option.get_addresses():
                bundle.mark_handled(option)



"""
An option handler that assigns addresses based on DUID from a SQLite database
"""
import codecs
import logging
import os
import sqlite3
import time
from ipaddress import IPv6Network, IPv6Address

from dhcpkit.ipv6.extensions.remote_id import RemoteIdOption
from dhcpkit.ipv6.options import ClientIdOption, InterfaceIdOption
from dhcpkit.ipv6.server.extensions.static_assignments import StaticAssignmentHandler, Assignment
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle

logger = logging.getLogger(__name__)


def build_sqlite() -> int:
    """
    Function to be called from the command line to convert a CSV based assignments file to a sqlite database.
    :return: exit code
    """
    import argparse
    import sys
    from dhcpkit.ipv6.server.extensions.static_assignments.csv import CSVStaticAssignmentHandler

    # Handle command line arguments
    parser = argparse.ArgumentParser(
        description="Assignments CSV to SQLite converter",
    )

    parser.add_argument("source", help="the source CSV file")
    parser.add_argument("destination", help="the destination SQLite file")
    parser.add_argument("-f", "--force", action="store_true", default=False, help="force removing old entries, even if "
                                                                                  "that means deleting more than 30%% "
                                                                                  "of the contents of the database")
    parser.add_argument("-v", "--verbosity", action="count", default=0, help="increase output verbosity")

    args = parser.parse_args()

    # Our logger is the root logger now
    global logger
    logger = logging.getLogger()

    # Don't filter on level in the root logger
    logger.setLevel(logging.NOTSET)

    # Output to sys.stdout
    stdout_handler = logging.StreamHandler(stream=sys.stdout)

    # Set level according to verbosity
    if args.verbosity >= 3:
        stdout_handler.setLevel(logging.DEBUG)
    elif args.verbosity == 2:
        stdout_handler.setLevel(logging.INFO)
    elif args.verbosity >= 1:
        stdout_handler.setLevel(logging.WARNING)
    else:
        stdout_handler.setLevel(logging.CRITICAL)

    logger.addHandler(stdout_handler)

    logger.info("Reading assignments from CSV file {}".format(args.source))
    csv_mtime = os.stat(args.source).st_mtime_ns
    logger.debug("CSV file modification time: {} ns".format(csv_mtime))
    assignments = CSVStaticAssignmentHandler.parse_csv_file(args.source)

    logger.info("Writing assignments to SQLite file {}".format(args.destination))
    db = sqlite3.connect(args.destination, isolation_level='IMMEDIATE')
    cur = db.cursor()
    cur.execute("CREATE TABLE IF NOT EXISTS assignments ("
                "id TEXT NOT NULL PRIMARY KEY, "
                "address TEXT, "
                "prefix TEXT, "
                "csv_mtime INT NOT NULL"
                ") WITHOUT ROWID")

    try:
        executed_in_transaction = 0
        for key, value in assignments:
            if executed_in_transaction == 0:
                # New transaction, check if we have a newer competing update process
                cur.execute("BEGIN IMMEDIATE")
                row = cur.execute("SELECT MAX(csv_mtime) FROM assignments").fetchone()
                if row[0] and row[0] > csv_mtime:
                    logger.critical("Update with newer CSV file detected, aborting")
                    return 1

            address = value.address and str(value.address) or None
            prefix = value.prefix and str(value.prefix) or None

            cur.execute("INSERT OR REPLACE INTO assignments (id, address, prefix, csv_mtime) VALUES (?, ?, ?, ?)",
                        (key, address, prefix, csv_mtime))

            executed_in_transaction += 1
            if executed_in_transaction >= 50:
                logger.debug("Interim commit to allow readers to access data")
                db.commit()
                time.sleep(0.05)
                executed_in_transaction = 0

        db.commit()

    except ValueError as e:
        logger.critical(e)
        return 1

    cur.execute("SELECT COUNT(1) FROM assignments")
    total_count = cur.fetchone()[0]
    logger.info("Database contains {} assignments".format(total_count))

    cur.execute("SELECT COUNT(1) FROM assignments WHERE csv_mtime=?", [csv_mtime])
    updated_count = cur.fetchone()[0]
    logger.info("Added/updated {} assignments".format(updated_count))

    safety_limit = total_count * 0.7
    do_delete = True
    if updated_count < safety_limit:
        if args.force:
            logger.warning("Removing old entries, despite that being >30% of total")
        else:
            logger.warning("Not removing old entries, would delete >30% of total")
            do_delete = False

    if do_delete:
        cur.execute("DELETE FROM assignments WHERE csv_mtime<?", [csv_mtime])
        logger.info("Deleted {} old assignments".format(cur.rowcount))

    db.commit()

    if not do_delete:
        # Signal that we didn't delete
        return 2
    else:
        # Normal exit
        return 0


class SqliteStaticAssignmentHandler(StaticAssignmentHandler):
    """
    Assign addresses and/or prefixes based on the contents of a Shelf file
    """

    def __init__(self, filename: str,
                 address_preferred_lifetime: int, address_valid_lifetime: int,
                 prefix_preferred_lifetime: int, prefix_valid_lifetime: int):
        """
        Initialise the mapping. This handler will respond to clients on responsible_for_links and assume that all
        addresses in the mapping are appropriate for on those links.

        :param filename: The filename containing the SQLite database
        """
        super().__init__(address_preferred_lifetime, address_valid_lifetime,
                         prefix_preferred_lifetime, prefix_valid_lifetime)

        self.sqlite_filename = filename
        self.db = None

    def __str__(self):
        return "{} from {}".format(self.__class__.__name__, self.sqlite_filename)

    def worker_init(self):
        """
        Open the SQLite database in each worker
        """
        logger.info("Opening SQLite database {}".format(self.sqlite_filename))
        self.db = sqlite3.connect(self.sqlite_filename, check_same_thread=False)

    def get_assignment(self, bundle: TransactionBundle) -> Assignment:
        """
        Look up the assignment based on DUID, Interface-ID of the relay closest to the client and Remote-ID of the
        relay closest to the client, in that order.

        :param bundle: The transaction bundle
        :return: The assignment, if any
        """
        # Gather all possible IDs
        possible_ids = []

        # Look up based on DUID
        duid_option = bundle.request.get_option_of_type(ClientIdOption)
        duid = 'duid:' + codecs.encode(duid_option.duid.save(), 'hex').decode('ascii')
        possible_ids.append(duid)

        # Look up based on Interface-ID
        interface_id_option = bundle.incoming_relay_messages[0].get_option_of_type(InterfaceIdOption)
        if interface_id_option:
            interface_id = 'interface-id:' + codecs.encode(interface_id_option.interface_id, 'hex').decode('ascii')
            possible_ids.append(interface_id)

        # Look up based on Remote-ID
        remote_id_option = bundle.incoming_relay_messages[0].get_option_of_type(RemoteIdOption)
        if remote_id_option:
            remote_id = 'remote-id:{}:{}'.format(remote_id_option.enterprise_number,
                                                 codecs.encode(remote_id_option.remote_id, 'hex').decode('ascii'))
            possible_ids.append(remote_id)

        # Search
        placeholders = ', '.join(['?'] * len(possible_ids))
        query = "SELECT address, prefix FROM assignments WHERE id IN (" + placeholders + ") ORDER BY id LIMIT 1"
        results = self.db.execute(query, possible_ids).fetchone()
        if results:
            address = results[0] and IPv6Address(results[0]) or None
            prefix = results[1] and IPv6Network(results[1]) or None

            return Assignment(address=address, prefix=prefix)

        # Nothing found
        return Assignment(address=None, prefix=None)



"""
The code to handle a message
"""
import logging
import multiprocessing

from dhcpkit.common.server.logging import DEBUG_HANDLING
from dhcpkit.ipv6.duids import DUID
from dhcpkit.ipv6.extensions.prefix_delegation import IAPDOption, IAPrefixOption
from dhcpkit.ipv6.messages import SolicitMessage, AdvertiseMessage, RequestMessage, \
    RenewMessage, RebindMessage, ReleaseMessage, InformationRequestMessage, DeclineMessage, ReplyMessage, \
    ConfirmMessage
from dhcpkit.ipv6.options import IANAOption, IATAOption, IAAddressOption, ClientIdOption, ServerIdOption, \
    StatusCodeOption, STATUS_USEMULTICAST
from dhcpkit.ipv6.server.extension_registry import server_extension_registry
from dhcpkit.ipv6.server.filters import Filter
from dhcpkit.ipv6.server.handlers import Handler, CannotRespondError, UseMulticastError
from dhcpkit.ipv6.server.handlers.client_id import ClientIdHandler
from dhcpkit.ipv6.server.handlers.interface_id import InterfaceIdOptionHandler
from dhcpkit.ipv6.server.handlers.rapid_commit import RapidCommitHandler
from dhcpkit.ipv6.server.handlers.server_id import ServerIdHandler, ForOtherServerError
from dhcpkit.ipv6.server.handlers.status_option import AddMissingStatusOptionHandler
from dhcpkit.ipv6.server.handlers.unanswered_ia import UnansweredIAOptionHandler
from dhcpkit.ipv6.server.handlers.unicast import RejectUnwantedUnicastHandler
from dhcpkit.ipv6.server.statistics import StatisticsSet
from dhcpkit.ipv6.server.transaction_bundle import TransactionBundle
from typing import List, Iterable

logger = logging.getLogger(__name__)


class MessageHandler:
    """
    Message processing class
    """

    def __init__(self, server_id: DUID, sub_filters: Iterable[Filter] = None, sub_handlers: Iterable[Handler] = None,
                 allow_rapid_commit: bool = False, rapid_commit_rejections: bool = False):
        self.server_id = server_id
        self.sub_filters = list(sub_filters or [])
        self.sub_handlers = list(sub_handlers or [])
        self.allow_rapid_commit = allow_rapid_commit
        self.rapid_commit_rejections = rapid_commit_rejections

        # Prepare static stuff
        self.setup_handlers = self.get_setup_handlers()
        self.cleanup_handlers = self.get_cleanup_handlers()

    def worker_init(self):
        """
        Separate initialisation that will be called in each worker process that is created. Things that can't be forked
        (think database connections etc) have to be initialised here.
        """
        logger.debug("Initialising MessageHandler in {}".format(multiprocessing.current_process().name))

        # Cascade to sub-filters and sub-handlers
        for sub_filter in self.sub_filters:
            sub_filter.worker_init()

        for sub_handler in self.sub_handlers:
            sub_handler.worker_init()

    def get_handlers(self, bundle: TransactionBundle) -> List[Handler]:
        """
        Get all handlers that are going to be applied to the request in the bundle.

        :param bundle: The transaction bundle
        :return: The list of handlers to apply
        """

        # Build the handlers list
        handlers = []
        """:type: [Handler]"""

        # Add setup handlers
        handlers += self.setup_handlers

        # Apply all sub-filters and collect their handlers. The idea behind this is that handlers on more-specific
        # filters take precedence over handlers on the outer filters.
        for sub_filter in self.sub_filters:
            handlers += sub_filter.get_handlers(bundle)

        # Now add our own handlers
        handlers += self.sub_handlers

        # Add cleanup handlers
        handlers += self.cleanup_handlers

        return handlers

    def get_setup_handlers(self) -> List[Handler]:
        """
        Build a list of setup handlers and cache it

        :return: The list of handlers
        """
        handlers = []
        """:type: [Handler]"""

        if self.allow_rapid_commit:
            # Rapid commit happens as the first thing in the post() stage
            handlers.append(RapidCommitHandler(self.rapid_commit_rejections))

        # These are mandatory
        handlers.append(ServerIdHandler(duid=self.server_id))
        handlers.append(ClientIdHandler())
        handlers.append(InterfaceIdOptionHandler())

        # Add setup handlers from extensions
        for extension_name, extension in server_extension_registry.items():
            create_setup_handlers = getattr(extension, 'create_setup_handlers', None)
            if create_setup_handlers:
                setup_handlers = create_setup_handlers()
                for setup_handler in setup_handlers:
                    logger.log(DEBUG_HANDLING, "Extension {} added {} to setup phase".format(
                        extension_name, setup_handler.__class__.__name__
                    ))
                handlers += setup_handlers

        return handlers

    @staticmethod
    def get_cleanup_handlers() -> List[Handler]:
        """
        Build a list of cleanup handlers and cache it

        :return: The list of handlers
        """
        handlers = []
        """:type: [Handler]"""

        # Reject unicast requests unless they have been explicitly permitted
        handlers.append(RejectUnwantedUnicastHandler())

        # Add cleanup handlers so they run last in the handling phase
        handlers.append(UnansweredIAOptionHandler())

        # Add cleanup handlers from extensions
        for extension_name, extension in server_extension_registry.items():
            create_cleanup_handlers = getattr(extension, 'create_cleanup_handlers', None)
            if create_cleanup_handlers:
                cleanup_handlers = create_cleanup_handlers()
                for cleanup_handler in cleanup_handlers:
                    logger.log(DEBUG_HANDLING, "Extension {} added {} to cleanup phase".format(
                        extension_name, cleanup_handler.__class__.__name__
                    ))
                handlers += cleanup_handlers

        # Confirm/Release/Decline messages always need a status
        handlers.append(AddMissingStatusOptionHandler())

        return handlers

    @staticmethod
    def init_response(bundle: TransactionBundle):
        """
        Create the message object in bundle.response

        :param bundle: The transaction bundle
        """
        # Start building the response
        if isinstance(bundle.request, SolicitMessage):
            bundle.response = AdvertiseMessage(bundle.request.transaction_id)

        elif isinstance(bundle.request, (RequestMessage, RenewMessage, RebindMessage,
                                         ReleaseMessage, DeclineMessage, InformationRequestMessage)):
            bundle.response = ReplyMessage(bundle.request.transaction_id)

        elif isinstance(bundle.request, ConfirmMessage):
            # Receipt of Confirm Messages: If [...] there were no addresses in any of the IAs sent by the client, the
            # server MUST NOT send a reply to the client.
            for option in bundle.request.get_options_of_type((IANAOption, IATAOption, IAPDOption)):
                if option.get_options_of_type((IAAddressOption, IAPrefixOption)):
                    # Found an address or prefix option
                    break
            else:
                # Not found: ignore request
                raise CannotRespondError

            bundle.response = ReplyMessage(bundle.request.transaction_id)

        else:
            logger.warning("Do not know how to reply to {}".format(type(bundle.request).__name__))
            raise CannotRespondError

        # Build the plain chain of relay reply messages
        bundle.create_outgoing_relay_messages()

    def construct_use_multicast_reply(self, bundle: TransactionBundle) -> ReplyMessage:
        """
        Construct a message signalling to the client that they should have used multicast.

        :param bundle: The transaction bundle containing the incoming request
        :return: The proper answer to tell a client to use multicast
        """
        # Make sure we only tell this to requests that came in over unicast
        if bundle.received_over_multicast:
            logger.error("Not telling client to use multicast, they already did...")
            return None

        return ReplyMessage(bundle.request.transaction_id, options=[
            bundle.request.get_option_of_type(ClientIdOption),
            ServerIdOption(duid=self.server_id),
            StatusCodeOption(STATUS_USEMULTICAST, "You cannot send requests directly to this server, "
                                                  "please use the proper multicast addresses")
        ])

    def handle(self, bundle: TransactionBundle, statistics: StatisticsSet):
        """
        The main dispatcher for incoming messages.

        :param bundle: The transaction bundle
        :param statistics: Container for shared memory with statistics counters
        :returns: The message to reply with
        """
        if not bundle.request:
            # Nothing to do...
            return None

        # Update the allow_rapid_commit flag
        bundle.allow_rapid_commit = self.allow_rapid_commit

        # Count the incoming message type
        statistics.count_message_in(bundle.request.message_type)

        # Log what we are doing (low-detail, so not DEBUG_HANDLING here)
        logger.debug("Handling {}".format(bundle))

        # Collect the handlers
        handlers = self.get_handlers(bundle)

        # Analyse pre
        for handler in handlers:
            # noinspection PyBroadException
            try:
                handler.analyse_pre(bundle)
            except:
                # Ignore all errors, analysis isn't that important
                pass

        try:
            # Pre-process the request
            for handler in handlers:
                handler.pre(bundle)

            # Init the response
            self.init_response(bundle)

            # Process the request
            for handler in handlers:
                logger.log(DEBUG_HANDLING, "Applying {}".format(handler))
                handler.handle(bundle)

            # Post-process the request
            for handler in handlers:
                handler.post(bundle)

        except ForOtherServerError:
            # Specific form of CannotRespondError that should have its own log message
            logger.debug("Message is for another server: ignoring")
            statistics.count_for_other_server()
            bundle.response = None

        except CannotRespondError:
            logger.debug("Cannot respond to this message: ignoring")
            statistics.count_do_not_respond()
            bundle.response = None

        except UseMulticastError:
            logger.debug("Unicast request received when multicast is required: informing client")
            statistics.count_use_multicast()
            bundle.response = self.construct_use_multicast_reply(bundle)

        # Analyse post
        for handler in handlers:
            # noinspection PyBroadException
            try:
                handler.analyse_post(bundle)
            except:
                # Ignore all errors, analysis isn't that important
                pass

        if bundle.response:
            logger.log(DEBUG_HANDLING, "Responding with {}".format(bundle.response.__class__.__name__))

            # Count the outgoing message type
            statistics.count_message_out(bundle.response.message_type)
        else:
            logger.log(DEBUG_HANDLING, "Not responding")



"""
Configuration elements for the NTP option handlers
"""
from dhcpkit.ipv6.extensions.ntp_suboption_registry import ntp_suboption_registry
from dhcpkit.ipv6.server.extensions.ntp import NTPServersOptionHandler
from dhcpkit.ipv6.server.handlers import HandlerFactory


class NTPServersOptionHandlerFactory(HandlerFactory):
    """
    Create the handler for NTP servers.
    """

    def clean_config_section(self):
        """
        Convert the data to the right types
        """
        converted_suboptions = {}

        for suboption_type, suboption_values in self.section.suboptions.items():
            suboption_class = ntp_suboption_registry.by_name[suboption_type]

            # See if this option supports loading from configuration
            datatype = suboption_class.config_datatype
            if not datatype:
                raise ValueError("{} cannot be used in the configuration".format(suboption_type))

            # Convert the values and store
            converted_suboption_values = [datatype(value) for value in suboption_values]
            converted_suboptions[suboption_type] = converted_suboption_values

        # And replace the string versions
        self.section.suboptions = converted_suboptions

    def validate_config_section(self):
        """
        Make sure the keys refer to actual NTP sub-options
        """
        for key in self.suboptions:
            # Lower-case the key so we have a canonical name
            key_lower = key.lower()

            # Check if the type exists
            if key_lower not in ntp_suboption_registry.by_name:
                raise ValueError("'{}' is not a known NTP server type".format(key))

    def create(self) -> NTPServersOptionHandler:
        """
        Create a handler of this class based on the configuration in the config section.

        :return: A handler object
        """

        # Create the real suboptions. I think that usually one suboption type will be used, but in case someone uses
        # different types let's sort them alphabetically to at least have consistency. Unfortunately we can't know the
        # original order in the config file anymore :(
        suboptions = []
        suboption_types = sorted(self.suboptions.keys())
        for suboption_type in suboption_types:
            suboption_class = ntp_suboption_registry.by_name[suboption_type]
            suboption_values = self.suboptions[suboption_type]
            for suboption_value in suboption_values:
                suboption = suboption_class(suboption_value)
                suboptions.append(suboption)

        return NTPServersOptionHandler(suboptions, always_send=self.always_send)



"""
Test the NTP option implementation
"""
import unittest
from ipaddress import IPv6Address

from dhcpkit.ipv6.extensions.ntp import NTPServersOption, NTPServerAddressSubOption, NTPMulticastAddressSubOption, \
    NTPServerFQDNSubOption, UnknownNTPSubOption, NTPSubOption
from tests.ipv6.options import test_option


class NTPServersOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('0038004b'
                                          '0001001020010db8000000000000000000000001'
                                          '00020010ff12000000000000000000000000abcd'
                                          '00030011') + b'\x03ntp\x08steffann\x02nl\x00' + \
                            bytes.fromhex('ffff000a') + b'RandomData'
        self.option_object = NTPServersOption(options=[
            NTPServerAddressSubOption(IPv6Address('2001:db8::1')),
            NTPMulticastAddressSubOption(IPv6Address('ff12::abcd')),
            NTPServerFQDNSubOption('ntp.steffann.nl'),
            UnknownNTPSubOption(65535, b'RandomData'),
        ])
        self.parse_option()

    def test_parse_wrong_type(self):
        with self.assertRaisesRegex(ValueError, 'does not contain NTPServerAddressSubOption data'):
            option = NTPServerAddressSubOption()
            option.load_from(b'00020010ff12000000000000000000000000abcd')

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'does not match the combined length'):
            NTPServersOption.parse(bytes.fromhex('003800130001001020010db8000000000000000000000001'))


class NTPSubOptionTestCase(test_option.OptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('ffff000a') + b'RandomData'
        self.option_object = UnknownNTPSubOption(65535, b'RandomData')
        self.parse_option()

    def parse_option(self):
        self.length, self.option = NTPSubOption.parse(self.option_bytes)
        self.assertIsInstance(self.option, NTPSubOption)
        self.option_class = type(self.option)

    def test_load_from_wrong_buffer(self):
        if issubclass(self.option_class, UnknownNTPSubOption):
            # UnknownNTPSubOption accepts any parseable buffer, no point in testing that one
            return

        super().test_load_from_wrong_buffer()


class UnknownNTPSubOptionTestCase(NTPSubOptionTestCase):
    def test_validate_suboption_type(self):
        self.check_unsigned_integer_property('suboption_type', size=16)

    def test_validate_suboption_data(self):
        self.option.suboption_data = b'AB'
        self.option.validate()

        self.option.suboption_data = 'ABC'
        with self.assertRaisesRegex(ValueError, 'sequence of bytes'):
            self.option.validate()


class NTPServerAddressSubOptionTestCase(NTPSubOptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('0001001020010db8000000000000000000000001')
        self.option_object = NTPServerAddressSubOption(IPv6Address('2001:db8::1'))
        self.parse_option()

    def test_config_datatype(self):
        value = NTPServerAddressSubOption.config_datatype('2001:db8::1')
        self.assertEqual(value, IPv6Address('2001:db8::1'))

        with self.assertRaisesRegex(ValueError, 'routable IPv6 address'):
            NTPServerAddressSubOption.config_datatype('::')

        with self.assertRaisesRegex(ValueError, 'routable IPv6 address'):
            NTPServerAddressSubOption.config_datatype('::1')

        with self.assertRaisesRegex(ValueError, 'routable IPv6 address'):
            NTPServerAddressSubOption.config_datatype('fe80::1')

        with self.assertRaisesRegex(ValueError, 'routable IPv6 address'):
            NTPServerAddressSubOption.config_datatype('ff02::1')

    def test_validate_address(self):
        self.option.address = bytes.fromhex('20010db8000000000000000000000001')
        with self.assertRaisesRegex(ValueError, 'routable IPv6 address'):
            self.option.validate()

        self.option.address = IPv6Address('2001:db8::1')
        self.option.validate()

        self.option.address = IPv6Address('::')
        with self.assertRaisesRegex(ValueError, 'routable IPv6 address'):
            self.option.validate()

        self.option.address = IPv6Address('::1')
        with self.assertRaisesRegex(ValueError, 'routable IPv6 address'):
            self.option.validate()

        self.option.address = IPv6Address('fe80::1')
        with self.assertRaisesRegex(ValueError, 'routable IPv6 address'):
            self.option.validate()

        self.option.address = IPv6Address('ff02::1')
        with self.assertRaisesRegex(ValueError, 'routable IPv6 address'):
            self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'must have length 16'):
            NTPServerAddressSubOption.parse(bytes.fromhex('0001000f'))

        with self.assertRaisesRegex(ValueError, 'must have length 16'):
            NTPServerAddressSubOption.parse(bytes.fromhex('00010011'))


class NTPMulticastAddressSubOptionTestCase(NTPSubOptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('00020010ff12000000000000000000000000abcd')
        self.option_object = NTPMulticastAddressSubOption(IPv6Address('ff12::abcd'))
        self.parse_option()

    def test_config_datatype(self):
        value = NTPMulticastAddressSubOption.config_datatype('ff02::1')
        self.assertEqual(value, IPv6Address('ff02::1'))

        with self.assertRaisesRegex(ValueError, 'multicast IPv6 address'):
            NTPMulticastAddressSubOption.config_datatype('::')

        with self.assertRaisesRegex(ValueError, 'multicast IPv6 address'):
            NTPMulticastAddressSubOption.config_datatype('::1')

        with self.assertRaisesRegex(ValueError, 'multicast IPv6 address'):
            NTPMulticastAddressSubOption.config_datatype('fe80::1')

        with self.assertRaisesRegex(ValueError, 'multicast IPv6 address'):
            NTPMulticastAddressSubOption.config_datatype('2001:db8::1')

    def test_validate_address(self):
        self.option.address = bytes.fromhex('20010db8000000000000000000000001')
        with self.assertRaisesRegex(ValueError, 'multicast IPv6 address'):
            self.option.validate()

        self.option.address = IPv6Address('ff02::1')
        self.option.validate()

        self.option.address = IPv6Address('::')
        with self.assertRaisesRegex(ValueError, 'multicast IPv6 address'):
            self.option.validate()

        self.option.address = IPv6Address('::1')
        with self.assertRaisesRegex(ValueError, 'multicast IPv6 address'):
            self.option.validate()

        self.option.address = IPv6Address('fe80::1')
        with self.assertRaisesRegex(ValueError, 'multicast IPv6 address'):
            self.option.validate()

        self.option.address = IPv6Address('2001:db8::1')
        with self.assertRaisesRegex(ValueError, 'multicast IPv6 address'):
            self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'must have length 16'):
            NTPMulticastAddressSubOption.parse(bytes.fromhex('0002000f'))

        with self.assertRaisesRegex(ValueError, 'must have length 16'):
            NTPMulticastAddressSubOption.parse(bytes.fromhex('00020011'))


class NTPServerFQDNSubOptionTestCase(NTPSubOptionTestCase):
    def setUp(self):
        self.option_bytes = bytes.fromhex('00030011') + b'\x03ntp\x08steffann\x02nl\x00'
        self.option_object = NTPServerFQDNSubOption('ntp.steffann.nl')
        self.parse_option()

    def test_config_datatype(self):
        value = NTPServerFQDNSubOption.config_datatype('ntp.steffann.nl')
        self.assertEqual(value, 'ntp.steffann.nl')

        with self.assertRaisesRegex(ValueError, 'letters, digits and hyphens'):
            NTPServerFQDNSubOption.config_datatype('something that is not a domain name')

        with self.assertRaisesRegex(ValueError, '1 to 63 characters long'):
            NTPServerFQDNSubOption.config_datatype('something..bad')

        with self.assertRaisesRegex(ValueError, '1 to 63 characters long'):
            NTPServerFQDNSubOption.config_datatype('steffann-steffann-steffann-steffann-'
                                                   'steffann-steffann-steffann-steffann.bad')

    def test_validate_fqdn(self):
        self.option.fqdn = ['steffann.nl']
        with self.assertRaisesRegex(ValueError, 'must be a string'):
            self.option.validate()

        self.option.fqdn = 'x' + '.x' * 127
        self.option.validate()

        self.option.fqdn = 'xx' + '.x' * 127
        with self.assertRaisesRegex(ValueError, 'must be 255 characters or less'):
            self.option.validate()

        self.option.fqdn = 'www.123456789012345678901234567890123456789012345678901234567890123.nl'
        self.option.validate()

        self.option.fqdn = 'www.1234567890123456789012345678901234567890123456789012345678901234.nl'
        with self.assertRaisesRegex(ValueError, 'must be 1 to 63 characters long'):
            self.option.validate()

    def test_bad_option_length(self):
        with self.assertRaisesRegex(ValueError, 'must end with a 0-length label'):
            NTPServerFQDNSubOption.parse(bytes.fromhex('0003000c') + b'\x08steffann\x02nl\x00')

        with self.assertRaisesRegex(ValueError, 'length does not match'):
            NTPServerFQDNSubOption.parse(bytes.fromhex('0003000e') + b'\x08steffann\x02nl\x00\x01')


if __name__ == '__main__':
    unittest.main()
